// DragUtils.cpp: implementation of the CDragUtils class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "tinycad.h"
#include "TinyCadDoc.h"
#include "Object.h"
#include "DragUtils.h"
#include "LineUtils.h"
#include "JunctionUtils.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDragUtils::CDragUtils(CTinyCadDoc* pDesign )
: m_j(pDesign)
{
	m_pDesign = pDesign;
	m_started = FALSE;

	// The stored begin data
	m_a = CPoint(0,0);
	m_b = CPoint(0,0);
}

CDragUtils::~CDragUtils()
{
	Clear();
}


void CDragUtils::Clear()
{
	// Now clear out our variables
	m_started = FALSE;
	m_discards.erase( m_discards.begin(), m_discards.end() );
	m_draggedWires.erase( m_draggedWires.begin(), m_draggedWires.end() );
}


// Find any wires that are horizontal or vertical and are joined to
// a wire in the same orientation.  Then use that to merge the wires
// together
//
void CDragUtils::Merge()
{
	// Look for any wires that are at a particular point...
	wireCollection::iterator it_wires = m_draggedWires.begin();
	while (it_wires != m_draggedWires.end())
	{
		CDrawingObject *pointer = (*it_wires).m_Object;
	
		// Is this a zero length wire?
		if (pointer->GetType() == xWire && m_discards.find( pointer ) == m_discards.end())
		{
			// Is this wire horizontal or vertical?
			bool vert = pointer->a.x == pointer->b.x;
			bool horiz = pointer->a.y == pointer->b.y;

			// if (vert || horiz)
			{
				CDrawingObject *merge_a = NULL;
				CDrawingObject *merge_b = NULL;

				bool is_junc_a = false;
				bool is_junc_b = false;

				// Look for any wire starting at this point
				// with the same orientation...
				// Or look for a junction at this point
				drawingIterator it = m_pDesign->GetDrawingBegin();
				while (it != m_pDesign->GetDrawingEnd()) 
				{
					CDrawingObject *test = *it;

					if (test->GetType() == xJunction)
					{
						if (test->a == pointer->a)
						{
							is_junc_a = true;
						}
						if (test->a == pointer->b)
						{
							is_junc_b = true;
						}
					}

					// Is this object not our comparision object  
					if (test != pointer && m_discards.find( test ) == m_discards.end())
					{
						// Is this a wire?
						if (!merge_a && test->GetType() == xWire && test->a != test->b)
						{
							// Is our pointer connected to this wire?
							CLineUtils l( test->a, test->b );
							CLineUtils p( pointer->a, pointer->b );

							double d;

							BOOL do_merge_a = l.IsPointOnLine( pointer->a, d);
							BOOL do_merge_b = l.IsPointOnLine( pointer->b, d);

							if (do_merge_a || do_merge_b)
							{
								bool test_vert = test->a.x == test->b.x;
								bool test_horiz = test->a.y == test->b.y;


								// Was is this a duplicate wire?  That is a wire that starts and ends
								// on the test wire
								if (do_merge_a && do_merge_b && l.GetLength() <= p.GetLength())
								{
									// Now make this wire discardable 
									// this will be picked up by the Clean() routine
									m_discards.insert( test );
								}
								else if (	(test_vert && test_vert == vert) 
									||	(test_horiz && test_horiz == horiz))
								{
									if (do_merge_a)
									{
										merge_a = test;
									}
									else
									{
										merge_b = test;
									}
								}
							}

						}
					}

					++ it;
				}

				if (merge_a && !is_junc_a)
				{
					CLineUtils l( pointer->a, pointer->b );
					double d;

					CPoint new_b = pointer->b;
					CPoint new_a = merge_a->a == pointer->a ? merge_a->b : merge_a->a;

					if (!l.IsPointOnLine(new_a,d))
					{
						pointer->a = new_a;
						pointer->b = new_b;
					}

					// Now make other wire discardable 
					// this will be picked up by the Clean() routine
					m_discards.insert( merge_a );
				}


				if (merge_b && !is_junc_b)
				{
					CLineUtils l( pointer->a, pointer->b );
					double d;

					CPoint new_a = pointer->a;
					CPoint new_b = merge_b->a == pointer->b ? merge_b->b : merge_b->a;

					if (!l.IsPointOnLine(new_b,d))
					{
						pointer->a = new_a;
						pointer->b = new_b;
					}
					// Now make other wire discardable 
					// this will be picked up by the Clean() routine
					m_discards.insert( merge_b );

				}
			}
		}

		++ it_wires;
	}
}

void CDragUtils::Clean()
{
	// We look at the lines we have played with to remove
	// any zero length wires we max have created by
	// this drag...
	// Find any wires attached to this pin and add them to the list
	wireCollection::iterator it_wires = m_draggedWires.begin();
	while (it_wires != m_draggedWires.end())
	{
		CDrawingObject *pointer = (*it_wires).m_Object;
	
		// Is this a zero length wire?
		if (pointer->GetType() == xWire && pointer->a == pointer->b)
		{
			// Yes, so delete it!
			m_discards.insert( pointer );
		}
		else
		{
			m_j.AddObjectToTodo( pointer );
		}

		++ it_wires;
	}

	// Now scan the discard list...
	discardCollection::iterator it_disc = m_discards.begin();
	while (it_disc != m_discards.end())
	{
		m_j.AddObjectToTodo( *it_disc );
		m_pDesign->Delete( *it_disc );
		++ it_disc;
	}
}


void CDragUtils::End( bool no_clean )
{
	// Clear out the discards list
	m_discards.erase( m_discards.begin(), m_discards.end() );

	if (!no_clean)
	{
		// Now determine which of these objects are still in the
		// drawing...
		wireCollection::iterator it_wires = m_draggedWires.begin();
		while (it_wires != m_draggedWires.end())
		{
			CDrawingObject *pointer = (*it_wires).m_Object;
			wireCollection::iterator prev_it_wires = it_wires ++;
			if (!m_pDesign->IsInDrawing( pointer ))
			{
				m_draggedWires.erase( prev_it_wires );
			}
		}

		// Merge any wires that are now straight!
		Merge();

		// Now discard any zero length wires
		Clean();

		// If we have made some changes then we had better re-check the
		// junctions...
		m_j.CheckTodoList( true );
	}

	Clear();
}


void CDragUtils::AddWireToCollection( dragWire n )
{
	// Now update this entry in the system...
	wireCollection::iterator it = m_draggedWires.find(n);
	if (it == m_draggedWires.end())
	{
		// New
		m_draggedWires.insert( n );
	
/*		TRACE("Adding %X Move_A(%d) = %X, Move_B(%d) = %X\n",
			n.m_Object,
			n.m_MoveA, n.m_Attached_LineA,
			n.m_MoveB, n.m_Attached_LineB );
*/
		// Now add any attached wires to this object...
		if (n.m_Object->GetType() == xWire || n.m_Object->GetType() == xBus)
		{
			AddAttachedObjects( n.m_Object, n.m_MoveA, n.m_MoveB );
		}

	}
	else
	{
		// Update
		if (n.m_MoveA && !(*it).m_MoveA)
		{
			(*it).m_MoveA |= n.m_MoveA;
			(*it).m_Attached_LineA = n.m_Attached_LineA;
			(*it).m_Distance_AlongA = n.m_Distance_AlongA;
		}
		if (n.m_MoveB && !(*it).m_MoveB)
		{
			(*it).m_MoveB |= n.m_MoveB;
			(*it).m_Attached_LineB = n.m_Attached_LineB;
			(*it).m_Distance_AlongB = n.m_Distance_AlongB;
		}

/*		TRACE("Updating %X Move_A(%d) = %X, Move_B(%d) = %X\n",
			(*it).m_Object,
			(*it).m_MoveA, (*it).m_Attached_LineA,
			(*it).m_MoveB, (*it).m_Attached_LineB );
*/
	}
}

// Add any wires that end at the point given to
// the drag list
void CDragUtils::AddWiresAtPoint( CPoint hs )
{
	// Find any wires attached to this pin and add them to the list
	drawingIterator it = m_pDesign->GetDrawingBegin();
	while (it != m_pDesign->GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;

		// Only search unselected objects
		if (!m_pDesign->IsSelected( pointer ))
		{
			if ((pointer->GetType()==xWire || pointer->GetType()==xBus) 
				&& (pointer->a == hs || pointer->b==hs)) 
			{
				// Build the undo list
				m_pDesign->MarkChangeForUndo( pointer );
				AddWireToCollection( dragWire(pointer, pointer->a == hs, pointer->b == hs) );
			}
		}

		++ it;
	}
}

// Given a wire, add any objects that are attached to it...
void CDragUtils::AddAttachedObjects( CDrawingObject *wire, BOOL wire_move_a, BOOL wire_move_b )
{
	CLineUtils l( wire->a, wire->b );


	// Find any wires attached to this wire and add them in...
	drawingIterator it = m_pDesign->GetDrawingBegin();
	while (it != m_pDesign->GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;

		// Is this the incoming wire?
		if (pointer == wire)
		{
			++ it;
			continue;
		}

		switch (pointer->GetType())
		{
		case xBus:
		case xWire:
			// Is this wire attached at either end to this
			// wire?
			double distance_along_a = 0;
			double distance_along_b = 0;
			BOOL move_a = FALSE;
			BOOL move_b = FALSE;
			
			move_a = l.IsPointOnLine( pointer->a, distance_along_a );
			move_b = l.IsPointOnLine( pointer->b, distance_along_b );

			// Does this object require moving?
			if (move_a || move_b)
			{
				// Build our drag wire
				bool use = false;
				dragWire n( pointer, move_a, move_b );

				if (move_a)
				{
					if (	(wire_move_a && distance_along_a < 1.0)
						||  (wire_move_b && distance_along_a > 0))
					{
						use |= true;
					}

					n.m_Distance_AlongA = distance_along_a;
					n.m_Attached_LineA = wire;
				}
				if (move_b)
				{
					if (	(wire_move_a && distance_along_b < 1.0)
						||  (wire_move_b && distance_along_b > 0))
					{
						use |= true;
					}

					n.m_Distance_AlongB = distance_along_b;
					n.m_Attached_LineB = wire;
				}

				// Build the undo list
				if (use)
				{
					m_pDesign->MarkChangeForUndo( n.m_Object );
					AddWireToCollection( n );
				}
			}
			break;
		}
		
		++ it;
	}
}

// Move any of the objects that are attached to this object...
//
void CDragUtils::MoveAttachedObjects( CDrawingObject *wire )
{
	CLineUtils l( wire->a, wire->b );

	wireCollection::iterator it_wires = m_draggedWires.begin();
	while (it_wires != m_draggedWires.end())
	{
		BOOL Updated = FALSE;
		CPoint move_by;

		if (!(*it_wires).m_Done_A && (*it_wires).m_MoveA && (*it_wires).m_Attached_LineA == wire) 
		{
			Updated = TRUE;
			CPoint r = l.GetPointOnLine( (*it_wires).m_Distance_AlongA, &m_pDesign->m_snap );
			move_by = (*it_wires).m_Object->a - r;
			(*it_wires).m_Object->a = r;
			(*it_wires).m_Done_A = TRUE; 

//			TRACE("Moving %X on A at %lg by %d,%d from %X\n", (*it_wires).m_Object,(*it_wires).m_Distance_AlongA,r.x,r.y, wire );
		}

		if (!(*it_wires).m_Done_B && (*it_wires).m_MoveB && (*it_wires).m_Attached_LineB == wire) 
		{
			Updated = TRUE;
			CPoint r = l.GetPointOnLine( (*it_wires).m_Distance_AlongB, &m_pDesign->m_snap );
			move_by = (*it_wires).m_Object->b - r;
			(*it_wires).m_Object->b = r;
			(*it_wires).m_Done_B = TRUE; 

//			TRACE("Moving %X on B at %lg by %d,%d from %X\n", (*it_wires).m_Object,(*it_wires).m_Distance_AlongB,r.x,r.y, wire );
		}


		if (Updated)
		{
			// (*it_wires).Lock();
			(*it_wires).m_Object->Display();

			// Now move any connected wires...
			MoveAttachedObjects( (*it_wires).m_Object );
		}

		++ it_wires;
	}
}

void CDragUtils::Begin( CPoint ai, CPoint bi )
{
	Clear();
	m_a = ai;
	m_b = bi;
}

void CDragUtils::Begin()
{
	m_started = TRUE;

	// Search for Methods which might be attached to wires via pins
	m_draggedWires.erase( m_draggedWires.begin(), m_draggedWires.end() );

	selectIterator it = m_pDesign->GetSelectBegin();
	while ( it != m_pDesign->GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;
		m_pDesign->MarkChangeForUndo( obj );

		switch (obj->GetType())
		{
		case xPower:
			// Are there any wires connected to this point?
			AddWiresAtPoint( obj->a );
			break;
		case xMethodEx3:
			{
				CDrawMethod *theMethod = static_cast<CDrawMethod*>((CDrawMethod*)obj);

				CPoint tr;
				drawingCollection method;
				theMethod->ExtractSymbol( tr, method );
				m_pDesign->SetPart(theMethod->GetSubPart());
				
				// Search the symbol for pins
				drawingIterator it = method.begin();
				while (it != method.end()) 
				{
					CDrawingObject *MethodPtr = (*it);
					CDrawPin* thePin = static_cast<CDrawPin*>((CDrawPin*)MethodPtr);

					// If it is a pin then search for wires attached
					if (MethodPtr->GetType()==xPinEx && !(thePin->IsInvisible())) 
					{
						CPoint hs = thePin->GetPosition(theMethod);
						AddWiresAtPoint( hs );
					}

					++ it;
				}
			}
		}

		++ it;
	}

	// Search for wires which might need one point fixed in place
	BOOL MoveA,MoveB;
	it = m_pDesign->GetSelectBegin();
	while ( it != m_pDesign->GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;
		MoveA=FALSE;
		MoveB=FALSE;

		// Is this a wire which needs to be draged?
		if (obj->GetType()==xWire || obj->GetType()==xBus) 
		{
			if (   obj->a.x>=min(m_a.x,m_b.x) && obj->a.x<=max(m_a.x,m_b.x)
			    && obj->a.y>=min(m_a.y,m_b.y) && obj->a.y<=max(m_a.y,m_b.y) )
				MoveA = TRUE;

			if (   obj->b.x>=min(m_a.x,m_b.x) && obj->b.x<=max(m_a.x,m_b.x)
			    && obj->b.y>=min(m_a.y,m_b.y) && obj->b.y<=max(m_a.y,m_b.y) )
				MoveB = TRUE;

			// If neither point is inside this, then
			// choose to move both...
			if (!MoveA && !MoveB)
			{
				MoveA = TRUE;
				MoveB = TRUE;
				AddWiresAtPoint( obj->a );
				AddWiresAtPoint( obj->b );
			}

			AddWireToCollection( dragWire( obj, MoveA, MoveB ) );
		}

		++ it;
	}
}

// Call "display" on all of the wires in the dragged list
void CDragUtils::DisplayDraggedWires()
{
	wireCollection::iterator it_wires = m_draggedWires.begin();
	while (it_wires != m_draggedWires.end())
	{
		m_j.AddObjectToTodo( (*it_wires).m_Object );
		(*it_wires).m_Object->Display();
		(*it_wires).m_Done_A = FALSE;
		(*it_wires).m_Done_B = FALSE;
		++ it_wires;
	}
}

void CDragUtils::MergeLinePoint( CDrawingObject *p )
{
	Clear();
	m_draggedWires.insert( dragWire(p, true, true) );
	End( false );
}


void CDragUtils::Drag( CPoint r )
{
	// Have we already started?
	if (!m_started)
	{
		Begin();
	}

	// Is there anything to do?
	if (r == CPoint(0,0))
	{
		return;
	}


	// With drag only move the points of a wire inside the update region
	selectIterator it = m_pDesign->GetSelectBegin();
	while ( it != m_pDesign->GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;

		bool defer = m_draggedWires.find(dragWire(obj)) != m_draggedWires.end();
		if (!defer)
		{
			obj->Display();

			m_j.AddObjectToTodo(obj);

			// Move the object
			obj->a += r;
			obj->b += r;

			m_j.AddObjectToTodo(obj);

			obj->Display();
		}
		
		++ it;
	}

	DisplayDraggedWires();

	wireCollection::iterator it_wires = m_draggedWires.begin();
	while (it_wires != m_draggedWires.end())
	{
		BOOL moved = FALSE;

		if ((*it_wires).m_MoveA && (*it_wires).m_Attached_LineA == NULL) 
		{
			(*it_wires).m_Done_A = TRUE;
			(*it_wires).m_Object->a += r;

			moved |= TRUE;
//			TRACE("Moving %X on A by %d,%d\n", (*it_wires).m_Object,r.x,r.y);
		}

		if ((*it_wires).m_MoveB && (*it_wires).m_Attached_LineB == NULL) 
		{
			(*it_wires).m_Done_B = TRUE;
			(*it_wires).m_Object->b += r;

			moved |= TRUE;
//			TRACE("Moving %X on B by %d,%d\n", (*it_wires).m_Object,r.x,r.y);
		}

		// (*it_wires).Lock();
	
		// Now move any attached items to this..
		if (moved)
		{
			MoveAttachedObjects( (*it_wires).m_Object );
		}

		++ it_wires;
	}

	DisplayDraggedWires();

	m_a += r;
	m_b += r;

	// Now check for junctions
	m_j.CheckTodoList( false );
}



void CDragUtils::Move( CPoint r )
{
	// Have we already started?
	if (!m_started)
	{
		Begin();
	}

	// Is there anything to do?
	if (r == CPoint(0,0))
	{
		return;
	}

	// Move all of the objects without dragging
	selectIterator it = m_pDesign->GetSelectBegin();
	while ( it != m_pDesign->GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;
		obj->Display();

		m_j.AddObjectToTodo(obj);

		// Move the object
		obj->a += r;
		obj->b += r;

		m_j.AddObjectToTodo(obj);

		obj->Display();

		++ it;
	}

	m_a += r;
	m_b += r;

	// Now check for junctions
	m_j.CheckTodoList( false );
}
