/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "LineUtils.h"
#include <math.h>
#include "JunctionUtils.h"


////// The Line Class //////


CDrawLine::CDrawLine(CTinyCadDoc *pDesign, ObjType NewType)
: CDrawingObject( pDesign )
, m_drag_utils_a( pDesign )
, m_drag_utils_b( pDesign )
{
	has_placed = FALSE;
	m_segment=1;
	xtype=NewType;
	is_stuck = FALSE;
	is_junction = FALSE;
	re_edit = FALSE;

	switch (xtype) 
	{
		case xBus:
			Style = fLINE;
			break;
		case xLineEx2:
			Style = m_pDesign->GetOptions()->GetCurrentStyle();
			break;
		case xDash:
			Style = fDASH;
			break;
		case xWire:
			Style = fLINE;
			break;
		default:
			Style = fLINE;
	}	
}


void CDrawLine::TagResources()
{
   m_pDesign->GetOptions()->TagStyle(Style);
}


const char* CDrawLine::GetXMLTag(ObjType t)
{
	switch (t) 
	{
	case xBus:
		return "BUS";
	case xWire:
		return "WIRE";
	default:
		return "LINE";
	}	
}


// Load and save to an XML file
void CDrawLine::SaveXML( CXMLWriter &xml )
{
	xml.addTag(GetXMLTag(xtype));

	switch (xtype) 
	{
	case xBus:
		break;
	case xWire:
		break;
	default:
		xml.addAttribute( "style", Style );
		break;
	}	

	xml.addAttribute( "a", CDPoint(a) );
	xml.addAttribute( "b", CDPoint(b) );
	xml.closeTag();
}

void CDrawLine::LoadXML( CXMLReader &xml )
{
	switch (xtype) 
	{
	case xBus:
		break;
	case xWire:
		break;
	default:
		xml.getAttribute( "style", Style );
		Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
		break;
	}	

	xml.getAttribute( "a", a );
	xml.getAttribute( "b", b );
}


// Load the line from a file
void CDrawLine::Load(CStream& archive )
{
 	a = ReadPoint(archive);
   	b = ReadPoint(archive);

	switch (xtype) {
		case xDash:
		case xLine:
			// Premote this type
			xtype = xLineEx2;
			break;
		case xLineEx:
			// Load the line thickness
			BYTE LineThickness;
			archive >> LineThickness;
			LineStyle l;
			l.Colour = cBLACK;
			l.Style = PS_SOLID;
			l.Thickness = LineThickness;
			Style = m_pDesign->GetOptions()->AddStyle(&l);
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			xtype = xLineEx2;
			break;
		case xLineEx2:
			archive >> Style;
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			break;
	}	
}


void CDrawLine::ToAngle()
{
  int FromX,FromY,FromC;

  switch (g_EditToolBar.m_LineEdit.GetAngle()) {
	case LINEBOX_90:
		if (a.x==b.x)
			g_EditToolBar.m_LineEdit.mode=2;
		if (a.y==b.y)
			g_EditToolBar.m_LineEdit.mode=1;
		if (g_EditToolBar.m_LineEdit.mode==0)
			g_EditToolBar.m_LineEdit.mode=1;
		break;
	case LINEBOX_45:
		g_EditToolBar.m_LineEdit.mode=0;
		// Which 45degree slope is it closest to?
		FromX=abs(a.x-b.x);
		FromY=abs(a.y-b.y);
		FromC=abs(FromX-FromY);
		// Is it closest to the X axis?
		if (FromX<FromY && FromX<FromC)
			b.x=a.x;
		// Is it closest to the Y axis?
		else if (FromY<FromC)
			b.y=a.y;
		// Must be closest to the 45 degree line
		else
			b.x=a.x+(b.x>a.x ? FromY : -FromY);
		break;
	default:
		g_EditToolBar.m_LineEdit.mode=0;
  }
}


BOOL CDrawLine::RButtonDown(CPoint p, CPoint s)
{
  BOOL r = !m_segment;


  // Remove any junctions we may have had in place...
  CJunctionUtils j(m_pDesign);
  j.CheckJunctionRequirement( a, false );
  j.CheckJunctionRequirement( b, false );

  Display();
 
  m_segment=1;
  a = b;
  has_placed = FALSE;
  is_stuck = FALSE;

  return r;
}


void CDrawLine::DblLButtonDown(CPoint p,CPoint s)
{
	// Second line..
	LButtonDown(p,s);

	// .. and end
	RButtonDown(p,s);
}


void CDrawLine::LButtonUp(CPoint p)
{
	// Only do this if this is the first
	// m_segment we have placed...
	if (has_placed)
	{
		return;
	}

	// If the user has made a significant movement,
	// then consider this a placement....
	int offset_x = a.x - b.x;
	int offset_y = a.y - b.y;

	if (offset_x < 0)
	{
		offset_x = -offset_x;
	}
	if (offset_y < 0)
	{
		offset_y = -offset_y;
	}

	if (offset_x > 10 || offset_y > 10)
	{
		// Ok, we can place
		if (!m_segment && !is_stuck)
		{
			// First line
			LButtonDown(p,p);
		}
		// Second line..
		LButtonDown(p,p);

		// .. and end
		RButtonDown(p,p);
	}
	else
	{
		has_placed = TRUE;
	}
}


void CDrawLine::NewOptions()
{
  if (xtype == xLineEx2) {
  	Display();
	Style = m_pDesign->GetOptions()->GetCurrentStyle();
	Display();
  }
}

int  CDrawLine::GetContextMenu()
{
	return IDR_LINETOOL_EDIT;
}

// For the context menu
void CDrawLine::ContextMenu( CPoint p, UINT id )
{
	if (m_segment)
	{
		// Determine the nearest end to the pointer
		CLineUtils l( a, b );
		double d;
		l.IsPointOnLine( p, d );

		switch (id)
		{
		case ID_CONTEXT_MAKEVERTICAL:
			if (a.y != b.y)
			{
				if (d < 0.5)
				{
					MoveField( 2, CPoint( a.x-b.x, 0) );
				}
				else
				{
					MoveField( 1, CPoint( b.x-a.x, 0) );
				}
			}
			break;
		case ID_CONTEXT_MAKEHORIZONTAL:
			if (a.x != b.x)
			{
				if (d < 0.5)
				{
					MoveField( 2, CPoint( 0,a.y-b.y) );
				}
				else
				{
					MoveField( 1, CPoint( 0,b.y-a.y) );
				}

			}
			break;
		}
	}
}


CString CDrawLine::GetName() const
{

  switch (xtype) {
	case xWire:
		return "Wire";
	case xBus:
		return "Bus";
	default:
		return "Line";
  }
}

int CDrawLine::getMenuID() 
{ 
  switch (xtype) {
	case xWire:
		return IDM_TOOLWIRE;
	case xBus:
		return IDM_TOOLBUS;
	default:
		return IDM_TOOLPOLYGON;
  }
}


void CDrawLine::LButtonDown(CPoint p, CPoint s)
{
  // New undo level for each placement...
  m_pDesign->BeginNewChangeSet();


  p = GetStickyPoint( p );

  if (m_segment) 
  {
	a=p;
	b=p;
	m_segment=!m_segment;

	  if (is_junction)
	  {
		  m_pDesign->AddNewJunction( a );
	  }
  } 
  else 
  {
	b=p;
	ToAngle();
	// Remove from screen
	Display();
	CPoint c=b;

	switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: b=CPoint(b.x,a.y);
			break;
		case 2: b=CPoint(a.x,b.y);
			break;
	}
	
	BOOL was_stuck = is_stuck;
	if (is_stuck && g_EditToolBar.m_LineEdit.mode > 0)
	{
		int old_is_junction = is_junction;
		is_junction = FALSE;

		// Place two lines
		is_stuck = FALSE;
		if (a != b)
		{
			// Store and then Check for merge & junctions
			m_drag_utils_a.MergeLinePoint( Store() );
		}
		a = b;
		b = c;

		is_junction = old_is_junction;
	}
	is_stuck = FALSE;
	if (a != b)
	{
	    // Store and then Check for merge & junctions
	    m_drag_utils_a.MergeLinePoint( Store() );
	}
	Display();	// Write to screen

	if (!was_stuck)
	{
		a=b;
		b=c;
		m_segment=0;
		ToAngle();
	}
	else
	{
		RButtonDown(p,s);
	}
  }
  Display();
}


void CDrawLine::Move(CPoint p, CPoint no_snap_p)
{
  if (!m_segment) 
  {
	  BOOL new_is_stuck = FALSE;
	  BOOL new_is_junction = FALSE;

	  // Are we a line to be snapped to pins?
	  if (xtype == xWire)
	  {
		  p = m_pDesign->GetStickyPoint(p, TRUE,TRUE,new_is_stuck,new_is_junction);
	  }


	  if (p != b)
	  {
		Display();
		is_stuck = new_is_stuck;
		is_junction = new_is_junction;
		b=p;
		ToAngle();
		Display();
	  }
  }
  else
  {
	  Display();
	  b = GetStickyPoint( p );
	  a = b;
	  Display();
  }
}


// Snap to a pin if we are drawing a wire...
//
CPoint CDrawLine::GetStickyPoint( CPoint q )
{
  // Are we a line to be snapped to pins?
  if (xtype != xWire)
  {
	  return q;
  }

  return m_pDesign->GetStickyPoint(q, TRUE,TRUE,is_stuck,is_junction);
}



ObjType CDrawLine::GetType()
{
  return xtype;
}

void CDrawLine::Display( BOOL erase )
{
	LineStyle *theStyle = m_pDesign->GetOptions()->GetStyle(Style);
	int width = max(2,theStyle->Thickness) * 2;


  CPoint r = a;
  if (!m_segment) 
  {
	  switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: 
			{
				r = CPoint(b.x,a.y);
				CLineUtils l(a, r);
				l.SplitForDisplay( m_pDesign,erase, width );
			}
			break;
		case 2:
			{
				r = CPoint(a.x,b.y);
				CLineUtils l(a, r);
				l.SplitForDisplay( m_pDesign,erase, width );
			}
			break;
	  }
  }

  CLineUtils l( r, b );
  l.SplitForDisplay( m_pDesign, erase, width );

 if (is_stuck)
 {
	 CRect r;
	 r.left = b.x - HIGHLIGHT_SIZE;
	 r.right = b.x + HIGHLIGHT_SIZE;
	 r.top = b.y - HIGHLIGHT_SIZE;
	 r.bottom = b.y + HIGHLIGHT_SIZE;
	 m_pDesign->InvalidateRect( r, erase, 3 );
 }
}


void CDrawLine::PaintHandles( Context &dc )
{
	dc.SelectBrush(cBLACK);
	dc.SelectPen(PS_SOLID,0,cBLACK);

	CRect r1(a.x-2,a.y-2,a.x+2,a.y+2);
	CRect r2(b.x-2,b.y-2,b.x+2,b.y+2);
	dc.Rectangle(r1);
	dc.Rectangle(r2);
}


// Move fields of this object about
int CDrawLine::IsInsideField(CPoint p)
{
	CLineUtils l(a,b);
	CRect r1(a.x-2,a.y-2,a.x+2,a.y+2);
	CRect r2(b.x-2,b.y-2,b.x+2,b.y+2);

	if (r1.PtInRect(p))
	{
		return 1;
	}
	if (r2.PtInRect(p))
	{
		return 2;
	}

	return -1;
}

BOOL CDrawLine::CanControlAspect()
{
	return FALSE;
}

void CDrawLine::MoveField(int w, CPoint r)
{
	Display();

	if (r != CPoint(0,0))
	{
		// Is the control key down?
		BOOL do_move = GetAsyncKeyState(VK_CONTROL) < 0;
		if (!m_pDesign->GetOptions()->GetAutoDrag())
		{
		  do_move = !do_move;
		}

		// Do we snap back onto the grid?
		if (m_pDesign->m_snap.GetGridSnap())
		{
			// Ok, force back onto the grid...
			if (w == 1)
			{
				CPoint q = m_pDesign->m_snap.Snap(a + r);
				r = q - a;
			}
			if (w == 2)
			{
				CPoint q = m_pDesign->m_snap.Snap(b + r);
				r = q - b;
			}

		}


		if (do_move)
		{
			CJunctionUtils j( m_pDesign );
			j.AddObjectToTodo( this );

			if (w == 1)
			{
				a += r;
			}
			if (w == 2)
			{
				b += r;
			}

			j.AddObjectToTodo( this );
			j.CheckTodoList( false );
		}
		else
		{
			if (w == 1)
			{
				m_drag_utils_a.Drag( r );
			}
			if (w == 2)
			{
				m_drag_utils_b.Drag( r );
			}
		}
	}

	Display();
}

int CDrawLine::SetCursorEdit( CPoint p )
{
	int s = IsInsideField( p );

	if (s != -1)
	{
		// Always display the same cursor
		s = 8;
	}
	else if (IsInside(p.x,p.x,p.y,p.y))
	{
		s = 11;
	}

	return s;
}



// Display the line on the screen!
void CDrawLine::Paint(Context &dc,paint_options options)
{
	switch (xtype)
	{
	case xWire:
		dc.SelectPen( PS_SOLID,1, m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::WIRE ), options );
		break;
	case xBus:
		dc.SelectPen( PS_SOLID,5, m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::BUS ), options );
		break;
	default:
		dc.SelectPen(m_pDesign->GetOptions()->GetStyle(Style), options);
		break;
	}
 
  dc.SetROP2(R2_COPYPEN);

  dc.MoveTo(a);

  if (!m_segment) {
	  switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: dc.LineTo(CPoint(b.x,a.y));
			break;
		case 2:	dc.LineTo(CPoint(a.x,b.y));
			break;
	  }
  }
  dc.LineTo(b);

  if (is_stuck)
  {
	  // Draw a nice circle to show the stickness...
	  dc.SelectBrush();
	  dc.SelectPen(PS_SOLID,5,cBOLD);
	  dc.SetROP2(R2_COPYPEN);
	  dc.Ellipse(CRect(b.x+HIGHLIGHT_SIZE,b.y+HIGHLIGHT_SIZE,b.x-HIGHLIGHT_SIZE,b.y-HIGHLIGHT_SIZE));

	  // Do we need a junction
	  if (is_junction)
	  {
	    int js=JUNCTION_SIZE;
		CPoint br,tl;
		br=CPoint(b.x+js+1,b.y+js+1);
		tl=CPoint(b.x-js,b.y-js);

		dc.SetROP2(R2_COPYPEN);

		dc.SelectPen(PS_SOLID,1,m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::JUNCTION) );
		dc.SelectBrush(m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::JUNCTION));
  	    dc.Ellipse(CRect(tl.x,tl.y,br.x,br.y));
	  }
  }

}


// Store the line in the drawing
CDrawingObject* CDrawLine::Store()
{
  // Only do this is we do not have zero length
  CDrawLine *NewObject;
  NewObject = new CDrawLine(m_pDesign);

  // Copy the details
  m_segment=1;
  *NewObject = *this;
  NewObject->is_junction = FALSE;
  NewObject->is_stuck = FALSE;

  m_pDesign->Add(NewObject);

  return NewObject;
}


void CDrawLine::BeginEdit(BOOL re_editi)
{
  re_edit = re_editi;
  if (re_edit)
  {
	  m_drag_utils_a.Begin( a,a );
	  m_drag_utils_b.Begin( b,b );
  }
  else
  {
	  m_segment=1;
	  g_EditToolBar.m_LineEdit.Open(m_pDesign,this);
	  g_EditToolBar.m_LineEdit.mode=0;

	  if (xtype == xLineEx2) 
	  {
		m_pDesign->GetOptions()->SetCurrentStyle(Style);
	  }
  }
 }

void CDrawLine::EndEdit()
{
  Display();
  if (!re_edit)
  {
	  g_EditToolBar.m_LineEdit.Close();
  }
  else
  {
	  m_drag_utils_a.End( true );
	  m_drag_utils_b.End( true );

	  CDragUtils q( m_pDesign );
	  q.MergeLinePoint( this );
  }
}

BOOL CDrawLine::IsInside(int left,int right,int top,int bottom)
{
	CLineUtils l(a,b);
	return l.IsInside(left,right,top,bottom);
}

double CDrawLine::DistanceFromPoint( CPoint p )
{
	CLineUtils l(a,b);
	CPoint d;
	return l.DistanceFromPoint(p, d );
}

