/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




// This handles the actual drawing of objects

#include "stdafx.h"
#include <math.h>
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "revision.h"




// The Constructor
CDrawMetaFile::CDrawMetaFile( CTinyCadDoc *pDesign )
: CDrawRectOutline( pDesign )
{
	m_segment=1;
	m_metafile = -1;
	a = CPoint(30,30);
	b = CPoint(130,130);
}


CDrawMetaFile::~CDrawMetaFile()
{
}

void CDrawMetaFile::setMetaFile( HENHMETAFILE data )
{
	HENHMETAFILE copy = CopyEnhMetaFile( data, NULL );

	ENHMETAHEADER hdr;
	GetEnhMetaFileHeader( data, sizeof(hdr), &hdr );

	int width = fabs((hdr.rclFrame.left - hdr.rclFrame.right) ) ;
	int height = fabs((hdr.rclFrame.bottom - hdr.rclFrame.top) ) ;

	b.x = fabs(b.y - a.y) * height / width;

	m_metafile = m_pDesign->GetOptions()->AddMetaFile( copy );
}

ObjType CDrawMetaFile::GetType()
{
  return xMetaFile;
}

CString CDrawMetaFile::GetName() const
{
  return "MetaFile";
}


void CDrawMetaFile::TagResources()
{
	m_pDesign->GetOptions()->TagMetaFile( m_metafile );
}


BOOL CDrawMetaFile::RButtonDown(CPoint p, CPoint s)
{
  BOOL r = !m_segment;

  Display();
  m_segment=1;
  
  return r;
}

// Load the rectange from a file
void CDrawMetaFile::Load(CStream& archive )
{
 // Read in the location
  a = ReadPoint(archive);
  b = ReadPoint(archive);
  archive >> m_metafile;
  m_metafile = m_pDesign->GetOptions()->GetNewMetaFileNumber(m_metafile);
}

const char* CDrawMetaFile::GetXMLTag()
{
	return "METAFILE";
}

// Load and save to an XML file
void CDrawMetaFile::SaveXML( CXMLWriter &xml )
{
	xml.addTag(GetXMLTag());

	xml.addAttribute( "a", CDPoint(a) );
	xml.addAttribute( "b", CDPoint(b) );
	xml.addAttribute( "id", m_metafile );

	xml.closeTag();

}

void CDrawMetaFile::LoadXML( CXMLReader &xml )
{
	xml.getAttribute( "a", a );
	xml.getAttribute( "b", b );
	xml.getAttribute( "id", m_metafile );
	m_metafile = m_pDesign->GetOptions()->GetNewMetaFileNumber(m_metafile);
}



void CDrawMetaFile::Paint(Context &dc,paint_options options)
{

  CPoint sma = a;
  CPoint smb = b;

  CRect rect(sma.x,sma.y,smb.x,smb.y);
  rect = dc.theTransform.Scale( rect );
  int rotmir = dc.theTransform.GetRotMir();

  // Do we need to restore the world transform?
  BOOL RestoreTransform = FALSE;

  // Save the current DC
  int dc_restore_point = dc.dc->SaveDC();

  dc.SelectDefault();

  // Does this device context support advanced transforms?
  if (SetGraphicsMode( dc.dc->m_hDC, GM_ADVANCED ) != 0)
  {
	  RestoreTransform = TRUE;

		XFORM xform;

	  // Normalise the rectangle
	  rect.NormalizeRect();
  
	  // Find out which way round the object goes
	  static const double PI = 4.0 * atan(1);
	  int offset_x = rect.left;
	  int offset_y = rect.top;
	  rect -= CPoint(offset_x, offset_y );
	  int fsx = 0;
	  int fsy = 0;
	  int msx = 0;
	  int msy = 0;
	  double rotation = 0;
	  switch (rotmir & 3) 
	  {
		case 0:		// Up
			rotation = 0;
			msx = rect.Width();
			break;
		case 1:		// Down
			rotation = PI;
			fsx = rect.Width();
			fsy = rect.Height();
			msy = rect.Height();
			break;
		case 2:		// Left
			rotation = (3.0*PI)/2.0;
			fsy = rect.Height();
			break;
		case 3:		// Right
			rotation = PI/2.0;
			fsx = rect.Width();
			msy = rect.Height();
			msx = rect.Width();
			break;
	  }


		xform.eM11 = cos(rotation);
		xform.eM12 = sin(rotation);
		xform.eM21 = -sin(rotation);
		xform.eM22 = cos(rotation);

		if ((rotmir &4)!=0)
		{
			if ((rotmir&3)<2)
			{
				xform.eM11 = -xform.eM11;
			}
			else
			{
				xform.eM12 = -xform.eM12;
			}
			fsx = msx;
			fsy = msy;
		}

		xform.eDx = offset_x + fsx;
		xform.eDy = offset_y + fsy;

		SetWorldTransform( dc.dc->m_hDC, &xform );
	}

  if ((rotmir & 3) > 1)
  {
	int width = rect.Width();
	int height = rect.Height();
	
	rect.right = rect.left + height;
	rect.bottom = rect.top + width;
  }


  PlayEnhMetaFile( dc.dc->m_hDC, m_pDesign->GetOptions()->GetMetaFile(m_metafile), &rect );

  if (RestoreTransform)
  {
	ModifyWorldTransform(dc.dc->m_hDC, NULL, MWT_IDENTITY );
	SetGraphicsMode( dc.dc->m_hDC, GM_COMPATIBLE );
  }

  // Now restore the DC
  dc.dc->RestoreDC( dc_restore_point );

}


// Store the rectangle in the drawing
CDrawingObject *CDrawMetaFile::Store()
{
  CDrawMetaFile *NewObject;

  NewObject = new CDrawMetaFile(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);
  
  m_segment=1;

  return NewObject;
}


