/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include <stdlib.h>
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "LineUtils.h"






////// The Pin Class //////


void CDrawPin::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(fPIN);
}


void CDrawPin::OldLoad(CStream &archive)
{

  a = ReadPoint(archive);
  archive >> which >> elec >> dir >> part >> str >> number >> show;
  DetermineSize();
  m_segment=0;

  // Sneekly hide the smallpin parameter in the which parameter....
  BOOL smallpin = which & 128;
  which &= ~128;
  number_pos = 0;
  length = (smallpin ? 20 : 30);
}

void CDrawPin::Load(CStream &archive)
{
  a = ReadPoint(archive);
  archive >> which >> elec >> dir >> part >> str >> number >> show >> length;
  DetermineSize();
  m_segment=0;

  if (length < 5)
  {
	  length = 20;
  }
  number_pos = 0;
}

const char* CDrawPin::GetXMLTag()
{
	return "PIN";
}

// Load and save to an XML file
void CDrawPin::SaveXML( CXMLWriter &xml )
{
	xml.addTag(GetXMLTag());
	xml.addAttribute( "pos", CDPoint(a) );
    xml.addAttribute( "which", which );
	xml.addAttribute( "elec", elec );
	xml.addAttribute( "direction", dir );
	xml.addAttribute( "part", part );
	xml.addAttribute( "number", number );
	xml.addAttribute( "show", show );
	xml.addAttribute( "length", length );
	xml.addAttribute( "number_pos", number_pos );

	xml.addChildData( str );

	xml.closeTag();
}

void CDrawPin::LoadXML( CXMLReader &xml )
{
    number_pos = 0;

	xml.getAttribute( "pos", a );
    xml.getAttribute( "which", which );
	xml.getAttribute( "elec", elec );
	xml.getAttribute( "direction", dir );
	xml.getAttribute( "part", part );
	xml.getAttribute( "number", number );
	xml.getAttribute( "show", show );
	xml.getAttribute( "length", length );
	xml.getAttribute( "number_pos", number_pos );

	xml.getChildData( str );

	m_segment=0;

	if (length < 5)
	{
		length = 20;
	}

	// Now determine it's size
	DetermineSize();
}


void CDrawPin::DetermineSize()
{
	// Is this pin visible?
	if (which == 4 || which == 5)
	{
		// Nope, so it has no size!
		b = a;
		return;
	}

	CPoint pa,pb,pc,pd,pta,ptb;
	int dr;
	DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, 0 );

	b = pta; 

	if ((show&1)!=0)
	{
	  CSize sz = m_pDesign->GetTextExtent( str, fPIN );
	  switch (dir)
	  {
	  case 0: // Up
		  b = CPoint( b.x, pta.y - sz.cx );
		  break;
	  case 1: // Down
		  b = pta;
		  break;
	  case 2: // Left
		  b = pta;
		  break;
	  case 3: // Right
		  b = CPoint( pta.x + sz.cx, b.y);
		  break;
	  }
	}

}

double CDrawPin::DistanceFromPoint( CPoint p )
{
  // Only inside if it is in the current part
  if (IsInvisible())
  {
	  return 100.0;
  }

  // Now get the line part and determine if the user has clicked
  // on it...
  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, 0 );
  CLineUtils l(a, pd);
  CPoint d;
  return l.DistanceFromPoint(p, d );
}

BOOL CDrawPin::IsInside(int left,int right,int top,int bottom)
{
  // Only inside if it is in the current part
  if (IsInvisible())
  {
	  return FALSE;
  }

  // Now get the line part and determine if the user has clicked
  // on it...
  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, 0 );
  CLineUtils l(a, pd);

  return l.IsInside(left,right,top,bottom );
}

BOOL CDrawPin::CanEdit()
{
  return TRUE;
}

ObjType CDrawPin::GetType()
{
  return xPinEx;
}


CString CDrawPin::GetName() const
{
  return "Pin";
}

// Look for a seach string in the object
CString CDrawPin::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = str;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? str : "";
}


// Construct a pin
CDrawPin::CDrawPin(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
  m_segment=1;
  which=0;
  dir=0;
  number="1";
  elec=0;
  str="";
  show=3;
  length = pDesign->GetOptions()->GetPinLength();
  part = m_pDesign->GetPart();
  number_pos = pDesign->GetOptions()->GetPinNumberPos();
}


CPoint CDrawPin::GetPosition(CDrawMethod *parent)
{
  CPoint r,p;
  CPoint bound = parent->GetTr();

  // Has this object been mirrored?
  if ((parent->GetRotate()&4)!=0)
	p = CPoint(-a.x + bound.x ,a.y);
  else
	p = a;

  switch ((parent->GetRotate())&3) {
	case 0:	// Up
		r.x = p.x + parent->a.x;
		r.y = p.y + parent->a.y;
		break;
	case 1: // Down
		r.x = p.x + parent->a.x;
		r.y = (bound.y - p.y) + parent->a.y;
		break;
	case 2: // Left
		r.x = p.y + parent->a.x;
		r.y = p.x + parent->a.y;
		break;
	case 3: // Right
		r.x = (bound.y - p.y) + parent->a.x;
		r.y = p.x + parent->a.y;
		break;
  }


  return r;
}


void CDrawPin::BeginEdit(BOOL re_edit)
{
  g_EditToolBar.m_PinEdit.Open(m_pDesign,this);
}

void CDrawPin::Move(CPoint p, CPoint no_snap_p)
{
  Display();
  a=p;
  DetermineSize();
  m_segment=0;
  Display();
}



// Determine weather or not this pin is visible
int CDrawPin::IsInvisible()
{
	int i = m_pDesign->GetPart();
 bool r = (part != i || (IsPower()  && !(m_pDesign->IsEditLibrary())) );

 return r;
}


#define PIN_UNSCALED	length
#define PIN_SIZE	PIN_UNSCALED

#define SIZING		((PIN_SIZE <= 20) ? 20 : 30)
#define DOT_SIZE	(SIZING/6)
#define LINE_SIZE	(PIN_SIZE-DOT_SIZE*2)
#define TEXT_SPACE	(SIZING/8)

void CDrawPin::DetermineLayout( CPoint &pa,CPoint &pb,CPoint &pc,CPoint &pd,CPoint &pta,CPoint &ptb, int &dr, int rotmir )
{
  int spacing;
  CPoint pos = a;
  CSize size;

  dr=3;

  if ((which&2)==0)
	spacing=PIN_SIZE+TEXT_SPACE;
  else
	spacing=PIN_SIZE+(DOT_SIZE*2)+TEXT_SPACE;

  // Find out which way round the pin object goes
  switch (dir) {
	case 0:		// Up
		pa=CPoint(pos.x,pos.y-LINE_SIZE);
		pb=CPoint(pos.x-DOT_SIZE,pos.y-PIN_SIZE);
		pc=CPoint(pos.x+DOT_SIZE+1,pos.y-PIN_SIZE);
		pd=CPoint(pos.x,pos.y-PIN_SIZE);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+size.cy/2,pos.y-spacing);
		ptb=CPoint(pos.x,pos.y-TEXT_SPACE - number_pos);
		dr=0;
		break;
	case 1:		// Down
		pa=CPoint(pos.x,pos.y+LINE_SIZE);
		pb=CPoint(pos.x-DOT_SIZE,pos.y+PIN_SIZE);
		pc=CPoint(pos.x+DOT_SIZE+1,pos.y+PIN_SIZE);
		pd=CPoint(pos.x,pos.y+PIN_SIZE);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+size.cy/2,pos.y+spacing+size.cx);
		size=m_pDesign->GetTextExtent(number, fPIN);
		ptb=CPoint(pos.x,pos.y+size.cx+TEXT_SPACE+ number_pos);
		dr=0;
		break;
	case 2:		// Left
		pa=CPoint(pos.x-LINE_SIZE,pos.y);
		pb=CPoint(pos.x-PIN_SIZE,pos.y-DOT_SIZE);
		pc=CPoint(pos.x-PIN_SIZE,pos.y+DOT_SIZE+1);
		pd=CPoint(pos.x-PIN_SIZE,pos.y);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x-size.cx-spacing,pos.y+size.cy/2);
		size=m_pDesign->GetTextExtent(number, fPIN);
		ptb=CPoint(pos.x-size.cx-TEXT_SPACE - number_pos,pos.y);
		break;
	case 3:		// Right
		pa=CPoint(pos.x+LINE_SIZE,pos.y);
		pb=CPoint(pos.x+PIN_SIZE,pos.y-DOT_SIZE);
		pc=CPoint(pos.x+PIN_SIZE,pos.y+DOT_SIZE+1);
		pd=CPoint(pos.x+PIN_SIZE,pos.y);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+spacing,pos.y+size.cy/2);
		ptb=CPoint(pos.x+TEXT_SPACE + number_pos,pos.y);
		break;
  }



  // Make corrections for bizzar rotations
  switch (dir)
  {
  case 2:
  case 3:
	    switch (rotmir)
		{
		case 1: // Down
		case 5: // Down & Mirror
		case 7: // Right
		case 3: // Right & Mirror
		  pta=CPoint(pta.x, pta.y - size.cy );
		  ptb=CPoint(ptb.x, ptb.y - size.cy );
		  break;
		}
  }

}


void CDrawPin::Display( BOOL erase )
{

  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, 0 );

	// Invalidate the symbol
	CRect r( a.x -2,a.y, pd.x + 2,pd.y);
	m_pDesign->InvalidateRect( r, erase, DOT_SIZE + 6 );

	// Invalidate the no-connect bit...
	CRect r2( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r2, erase, CONNECT_SIZE + 2 );


  if ((show&1)!=0)
  {
	  CSize sz = m_pDesign->GetTextExtent( str, fPIN );
	  if (dr < 2)
	  {
		  r = CRect( pta.x, pta.y, pta.x - sz.cy, pta.y - sz.cx);
	  }
	  else
	  {
		  r = CRect( pta.x, pta.y - sz.cy, pta.x + sz.cx, pta.y + sz.cy );
	  }
	  m_pDesign->InvalidateRect( r, erase, 10 );
  }
  if ((show&2)!=0)
  {
  	  CSize sz = m_pDesign->GetTextExtent( number, fPIN );
	  if (dr < 2)
	  {
		  r = CRect( ptb.x - sz.cy, ptb.y - sz.cy, ptb.x + sz.cy, ptb.y + sz.cx);		  
	  }
	  else
	  {
		  r = CRect( ptb.x, ptb.y - sz.cy, ptb.x + sz.cx, ptb.y + sz.cy );
	  }
	  
	  m_pDesign->InvalidateRect( r, erase, 10 );	
  }

}


// Display the pin item on the screen!
void CDrawPin::Paint(Context &dc,paint_options options)
{
  // Only display the pin if the it is in the current part
  if (IsInvisible() || (which == 5 && !(m_pDesign->IsEditLibrary())) )
	return;

  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, dc.theTransform.GetRotMir() );

  dc.SetROP2(R2_COPYPEN);
  dc.SetTextColor(m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::POWER));

  dc.SelectBrush();
  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(fPIN),dr);

  switch (options)
  {
  case draw_selected:
	    dc.SelectPen(PS_SOLID,1,cSELECT );
	  break;
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cPIN_CLK);
	  break;
  default:
	    dc.SelectPen(PS_SOLID,1,m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::PIN));
  }


  if ((which & 1)!=0 && which < 4 ) 
  {
	// Draw the dot
	dc.Ellipse(CRect(pb.x+(pa.x-pd.x),pb.y+(pa.y-pd.y),pc.x,pc.y));
  } 
  else 
  {
	dc.MoveTo(pa);
	dc.LineTo(pd);
  }

  // Draw the clock (triangle)
  if ((which & 2)!=0) 
  {
	dc.MoveTo(pb);
  	dc.LineTo(pc);
	dc.MoveTo(pb);
	dc.LineTo(CPoint(pd.x-(pa.x-pd.x),pd.y-(pa.y-pd.y)));
	dc.LineTo(pc);
  }



  // Draw the bits that all pins have in comman
  dc.MoveTo(a);
  dc.LineTo(pa);

  if ((show&1)!=0)
  {
	dc.TextOut(str,pta,options,dr);
  }

  if ((show&2)!=0)
	dc.TextOut(number,ptb,options,dr);

  // If this pin is a "not connected" then we must draw the NC cross on
  // the end of it....
  if (elec == 6)
  {
	  CPoint LineTo1=CPoint(a.x+CONNECT_SIZE,a.y+CONNECT_SIZE);
	  CPoint LineTo2=CPoint(a.x+CONNECT_SIZE,a.y-CONNECT_SIZE);

	  dc.MoveTo(CPoint(a.x-CONNECT_SIZE,a.y-CONNECT_SIZE));
	  dc.LineTo(CPoint(LineTo1.x+1,LineTo1.y));
	  dc.MoveTo(CPoint(a.x-CONNECT_SIZE,a.y+CONNECT_SIZE));
	  dc.LineTo(CPoint(LineTo2.x+1,LineTo2.y));
  }
}


void CDrawPin::EndEdit()
{
  RButtonDown(CPoint(0,0), CPoint(0,0));
  g_EditToolBar.m_PinEdit.Close();
}


void CDrawPin::LButtonDown(CPoint p, CPoint)
{
  // New undo level for each placement...
  m_pDesign->BeginNewChangeSet();

  Display();
  a=p;
  DetermineSize();
  Store();
  Display();	// Write to screen

  // Add 1 to the pin number (if possible)
  int hold=atoi(number);
  // Only increment if old value was a number
  if (hold!=0) {
	hold += m_pDesign->GetPinDir();
	if (hold>999 || hold<1)
		hold=1;
	_itoa(hold++,number.GetBuffer(17),10);
    number.ReleaseBuffer();
  }

  hold=str.FindOneOf("0123456789-");
  // Only increment if old value was a number
  if (hold!=-1) {
	char NewNum[17];
	_itoa(atoi(str.Mid(hold))+m_pDesign->GetNameDir(),NewNum,10);
	str = str.Left(hold)+NewNum;
  }


  g_EditToolBar.m_PinEdit.ReFocus();

  Display();
}


int CDrawPin::DoRotate(int olddir,int newdir)
{
  //     New rotation=> 2  3  4  
  //				    Current dir ..\/..
  int table[] = {	2, 3, 0, 		// 0 (Up)
			3, 2, 1,		// 1 (Down)
			1, 0, 3,		// 2 (Left)
			0, 1, 2,		// 3 (Right)
		};

  return table[(newdir-2) + olddir*3];
}



// Rotate this object about a point
void CDrawPin::Rotate(CPoint p,int ndir)
{
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  dir = DoRotate(dir,ndir);
}



// Store the line in the drawing
CDrawingObject* CDrawPin::Store()
{
  CDrawPin *NewObject;

  NewObject = new CDrawPin(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  return NewObject;
}

