/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "stdafx.h"
#include <set>
#include <map>

#include "TinyCadView.h"
#include "TinyCad.h"
#include "special.h"
#include "library.h"
#include "net.h"
#include "registry.h"
#include "diag.h"
#include "LineUtils.h"


// Construction/destruction

CNetList::CNetList()
{
	m_CurrentNet = 1;
}


CNetList::~CNetList()
{
}



// Get the next available netlist number...
int CNetList::GetNewNet()
{
	return m_CurrentNet++;
}



// Add a node to the tree (if necessary), return value of the netlist
int CNetList::Add(CNetListNode &ins)
{
	// Is this node already in the tree?
	int found = m_nodes[ ins.m_a ];
  
	// Has this node already been assigned a net-list index?
	if (ins.m_NetList == -1) 
	{
		// No, so we can add without checking for prior connections...
		if (found != 0) 
		{
			ins.m_NetList = found;
		}
		else
		{
			ins.m_NetList = GetNewNet();
			m_nodes[ins.m_a] = ins.m_NetList;
		}

		m_nets[ins.m_NetList].push_back( ins );

		return ins.m_NetList;
	} 
	else 
	{
		// If this node was already found, but with a different netlist
		// number, then first we must join the two netlists together
		if (found != 0 && found != ins.m_NetList) 
		{
			// The two nets must be joined
			int NewList = ins.m_NetList;
			int OldList = found;

			// Get the old list
			nodeVector::iterator OldNetList = m_nets[OldList].begin();

			// Make the old list's net-list index the same as the new list
			while (OldNetList != m_nets[OldList].end()) 
			{
				(*OldNetList).m_NetList = NewList;
				m_nodes[(*OldNetList).m_a] = NewList;
				++ OldNetList;
			}


			// Now concatenate the old lists onto the new list
			std::copy(m_nets[OldList].begin(), m_nets[OldList].end(), std::back_inserter(m_nets[NewList]));

			// Delete the old list
			netCollection::iterator it = m_nets.find(OldList);
			m_nets.erase( it );

			ins.m_NetList = NewList;
		}

		// Add this node to the netlist
		m_nets[ins.m_NetList].push_back( ins );
		m_nodes[ins.m_a] = ins.m_NetList;

		return ins.m_NetList;
	}
}

// Link together several netlists
//
// This is done by linking together:
//  nets that have nodes that have the same filename and same label name
//
// 
void CNetList::Link( linkCollection& nets )
{
	// Get rid of any old data
	m_CurrentNet = 1;
	m_nodes.erase( m_nodes.begin(), m_nodes.end() );
	m_nets.erase( m_nets.begin(), m_nets.end() );

	// Here is a list of known nodes and their netlist number
	typedef std::map<CString,int> stringCollection;
	stringCollection	labels;

	// Here is a list of known linking information
	typedef std::map<int,int> intCollection;
	typedef std::vector< intCollection > linkMap;
	linkMap	map;

	// Here is a map of netlists that are linked together
	intCollection	linked_netlists;

	// The linkmap will contain for each netlist a map
	// of old net number to new net number...
	map.resize( nets.size() );

	// Now scan the nodes to generate the superset of
	// labels and their linking information
	int index = 0;
	linkCollection::iterator i = nets.begin();
	while ( i != nets.end() )
	{
		CNetList &n = *i;

		// Scan each netlist
		netCollection::iterator ni = n.m_nets.begin();
		while (ni != n.m_nets.end())
		{
			nodeVector &v = (*ni).second;
			int old_netlist = (*ni).first;
			int new_netlist = 0;

			// Scan each node
			nodeVector::iterator vi = v.begin();
			while (vi != v.end())
			{
				CNetListNode &node = *vi;

				if (!node.m_label.IsEmpty())
				{
					// We only link at label level...
					CString label_name = node.m_filename + "::" + node.m_label;

					// Do we already have an entry for this label?
					stringCollection::iterator s = labels.find( label_name );
					if (s != labels.end())
					{
						int q = (*s).second;

						while (linked_netlists.find(q) != linked_netlists.end())
						{
							q = linked_netlists[ q ];
						}

						// Have we already assigned a netlist to this node?
						if (new_netlist != 0)
						{
							int z = new_netlist;
							while (linked_netlists.find(z) != linked_netlists.end())
							{
								int hold = linked_netlists[ z ];
								if (z != q)
								{
									linked_netlists[ z ] = q;
								}
								else
								{
									// No need to link a list to itself!
									linked_netlists.erase( q );
								}
								z = hold;
							}

							// We will have to link these two nodes together...
							if (new_netlist != q)
							{
								linked_netlists[ new_netlist ] = q;
							}
						}
						else
						{
							// Yes... so this node must become that netlist
							new_netlist = q;
						}
					}
					else
					{
						// No, so we must insert a new node..
						if (new_netlist == 0)
						{
							new_netlist = m_CurrentNet ++;
						}
						
						labels[ label_name ] = new_netlist;
					}

				}

				++ vi;
			}

			// Did we find a new netlist number for this node?
			if (new_netlist == 0)
			{
				new_netlist = m_CurrentNet ++;
			}

			// Now write it into the map
			map[ index ][ old_netlist ] = new_netlist;

			++ ni;
		}

		++ index;
		++ i;
	}


	// Now build a super net of all of the
	// netlists linked together...
	index = 0;
	i = nets.begin();
	while ( i != nets.end() )
	{
		CNetList &n = *i;

		// Scan each netlist
		netCollection::iterator ni = n.m_nets.begin();
		while (ni != n.m_nets.end())
		{
			nodeVector &v = (*ni).second;
			int old_netlist = (*ni).first;
			int new_netlist = map[ index ][ old_netlist ];

			// Check to see if these netlists are linked together...
			while (linked_netlists.find(new_netlist) != linked_netlists.end())
			{
				new_netlist = linked_netlists[ new_netlist ];
			}


			// Now concatenate the old lists onto the new list
			std::copy(v.begin(), v.end(), std::back_inserter(m_nets[new_netlist]));

			++ ni;
		}

		++ index;
		++ i;
	}


}


// Perform the work of making a netlist from this design...
void CNetList::MakeNet( CTinyCadMultiDoc *pDesign )
{
	linkCollection nets;
	nets.resize( pDesign->GetNumberOfSheets() );

	// Generate a netlist for every sheet in this design
	for (int i = 0; i < pDesign->GetNumberOfSheets(); i++)
	{
		nets[i].MakeNet( pDesign->GetPathName(), pDesign->GetSheet( i ) );
	}

	// Now link the nets together
	Link( nets );
}




// Connections between nets (the nodes) are at:
// .. the end of wires
// .. the point of pins
// .. the point of power
// .. the point of labels
// .. the point of no-connects

// To generate a netlist:
// 1. Search design for nodes.
//  - Insert nodes into node tree and either assign net list to node or gernerate a new net list
//  - If wire joins two different netlists, then join the two netlists together
// 2. Find Junctions
//  - Find which wires junctions lie on and connect their netlists
// 3. Find Labels
//  - Find which wire each label lies on and connect their netlists
// Perform the work of making a netlist from this design...

// Perform the work of making a netlist from a single sheet in this design...
void CNetList::MakeNet( const char *filename, CTinyCadDoc *pDesign )
{
  // Get rid of any old data
  m_CurrentNet = 1;
  m_nodes.erase( m_nodes.begin(), m_nodes.end() );
  m_nets.erase( m_nets.begin(), m_nets.end() );

  // Here is some temporary data for this function
  typedef std::map<CString,int> stringCollection;
  stringCollection	Powers;
  stringCollection	Connected;

  // Search for nodes, and build the node tree
  drawingIterator it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;
	stringCollection::iterator found;
	int hold;
	CPoint tr;

	switch (ObjPtr->GetType()) 
	{
		case xMethodEx3:
			#define thePin ((CDrawPin*)pointer)
			#define theMethod ((CDrawMethod*)ObjPtr)
			hold = pDesign->GetPart();
			pDesign->SetPart(theMethod->GetSubPart());

			{
				drawingCollection method;
				((CDrawMethod *)ObjPtr)->ExtractSymbol(tr,method);

				drawingIterator it = method.begin();
				while ( it != method.end() ) 
				{
					CDrawingObject *pointer = *it;

					if (pointer->GetType()==xPinEx && !(thePin->IsPower()) && !(thePin->IsInvisible())) 
					{
						CNetListNode n(filename, thePin,thePin->GetPosition(theMethod));
						n.m_reference = theMethod->GetRef();
						n.m_pin = thePin->GetNumber();
						n.m_pMethod = theMethod;
						Add(n);
					}

					++ it;
				}
				pDesign->SetPart(hold);

				// Has this symbol had it's power connected?
				if (Connected.find(theMethod->GetRef()) == Connected.end()) 
				{
					Connected[ theMethod->GetRef() ] = TRUE;

					drawingCollection method;
					((CDrawMethod *)ObjPtr)->ExtractSymbol(tr,method);
					drawingIterator it = method.begin();
					while (it!=method.end()) 
					{
						CDrawingObject *pointer = *it;

						if (pointer->GetType()==xPinEx && thePin->IsPower()) 
						{
							CNetListNode n(filename, thePin,thePin->GetPosition(theMethod) );
							n.m_reference = theMethod->GetRef();
							n.m_pin = thePin->GetNumber();
							n.m_pMethod = theMethod;


							// Look up the netlist this power belongs to
							found = Powers.find( thePin->GetPinName() );
							if (found != Powers.end())
								n.m_NetList = (*found).second;
							hold = Add(n);
							if (found == Powers.end())
								Powers[thePin->GetPinName()] = hold;
						}

						++ it;
					}
				}
			}			
			break;
		case xNoConnect:
			Add(CNetListNode(filename, ObjPtr,ObjPtr->a));			
			break;
		case xJunction:
			Add(CNetListNode(filename, ObjPtr,ObjPtr->a));
			break;			
		case xPower:
			{
			CNetListNode n(filename, ObjPtr,ObjPtr->a);
			n.m_label = ((CDrawPower *)ObjPtr)->GetValue();

			// Does this power item exist?
			found = Powers.find(((CDrawPower *)ObjPtr)->GetValue());
			if (found != Powers.end())
				n.m_NetList = (*found).second;
			hold = Add(n);
			if (found == Powers.end())
				Powers[((CDrawPower *)ObjPtr)->GetValue()] = hold;
			}
			break;
		case xWire:
			{
				CNetListNode n(filename, ObjPtr,ObjPtr->a);
				hold = Add(n);
			}
			{
				CNetListNode n(filename, ObjPtr,ObjPtr->b);
				n.m_NetList = hold;
				Add(n);
			}
			break;
	}
	
	++ it;
  }

 

  // Search for junctions and connect together
  it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;

	// Search for junctions
	if (ObjPtr->GetType() == xJunction) 
	{
		// Find out which netlist was assigned to this point
		CPoint a = ObjPtr->a;
		int NetNumber = m_nodes[ a ];

		// Look for wires which cross this junction
		drawingIterator search_it = pDesign->GetDrawingBegin();
		while (search_it != pDesign->GetDrawingEnd()) 
		{
			CDrawingObject *search = *search_it;

			// Find the wires
			// If the wire has an end at this junction then it is already connected
			if (search->GetType()==xWire 
			 && search->a!=a && search->b!=a)
			{
				// Is this point on this wire?
				CLineUtils l( search->a, search->b );
				double distance_along_a;

				if (l.IsPointOnLine( a, distance_along_a ))
				{
					CNetListNode n(filename, NULL,search->a);
					n.m_NetList = NetNumber;
					NetNumber = Add(n);
				}
			}

			++ search_it;
		}
	}
	
	++ it;
  }

  // Search for labels and connect to their respective lines
  stringCollection labels;
  it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;

	// Search for junctions
	if (ObjPtr->GetType() == xLabelEx2) 
	{
		CPoint a = ObjPtr->a;

		// Search for a wire this label is connect to
		// Only atempt to connect to a single wire
		drawingIterator search_it = pDesign->GetDrawingBegin();
		while (search_it != pDesign->GetDrawingEnd()) 
		{
			CDrawingObject *search = *search_it;
			if (search->GetType()==xWire && search->IsInside(a.x,a.x,a.y,a.y)) 
			{
				a = search->a;
				break;
			}
			
			++ search_it;
		}

		// Look up this label
		CNetListNode n(filename, ObjPtr,a);
		n.m_label = ((CDrawText *)ObjPtr)->GetValue();

		// Has this label already been assigned a netlist?
		stringCollection::iterator found = labels.find(((CDrawText *)ObjPtr)->GetValue());
		if (found!=labels.end()) 
		{
			n.m_NetList = (*found).second;
			n.m_label = "";
		}

		int hold = Add(n);

		// If there was no netlist write it back...
		if (found == labels.end())
			labels[ ((CDrawText *)ObjPtr)->GetValue()] = hold;
	}
	
	++ it;
  }  

  // Our work with the nodes map is complete, so we can discard
  // it now...
  m_nodes.erase( m_nodes.begin(), m_nodes.end() );

}


// Create netlist and output as a SPICE file
void CNetList::WriteNetListFile( int type, CTinyCadMultiDoc *pDesign, const char *filename )
{
	switch (type)
	{
	case 1:
		WriteNetListFilePADS( pDesign, filename );
		break;
	case 2:
		WriteNetListFileEagle( pDesign, filename );
		break;
	case 3:
		WriteNetListFileProtel( pDesign, filename );
		break;
	default:
		WriteNetListFileTinyCAD( pDesign, filename );
		break;
	}
}

// Create netlist and output as a Potel PCB script
void CNetList::WriteNetListFileProtel( CTinyCadMultiDoc *pDesign, const char *filename )
{
  FILE *theFile = fopen(filename,"w");
  if (theFile == NULL) 
  {
	Message(IDS_CANNOTOPEN);
	return;
  }

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

	// Get the net list
	MakeNet( pDesign );


  	// Generate a component for every sheet in this design
	for (int i = 0; i < pDesign->GetNumberOfSheets(); i++)
	{
		CString last = "";

		drawingIterator it = pDesign->GetSheet(i)->GetDrawingBegin();
		while (it != pDesign->GetSheet(i)->GetDrawingEnd()) 
		{
			CDrawingObject *pointer = *it;

			if (pointer->GetType() == xMethodEx3 && ((CDrawMethod*)pointer)->GetRef()!=last) 
			{
				CDrawMethod *pMethod = static_cast<CDrawMethod *>(pointer);
				CString Ref  = pMethod->GetRef();
				CString Package = "This part has no 'Package' attribute";
				CString Name = pMethod->GetName();

				for (int i = 2; i < pMethod->GetFieldCount(); i++)
				{
					if (pMethod->GetFieldName(i).CompareNoCase("package") == 0)
					{
						Package = pMethod->GetField(i);
					}
				}

				fprintf(theFile,"[\n%s\n%s\n%s\n\n\n\n]\n", 
				Ref, Package, Name);

				last = ((CDrawMethod *)pointer)->GetRef();
			}

			++ it;
		}
	}


  int Label = 0;

  netCollection::iterator nit = m_nets.begin();

	while (nit != m_nets.end()) 
	{
		nodeVector::iterator nv_it = (*nit).second.begin();

		CString theLine,theLabel;

		if (nv_it != (*nit).second.end()) 
		{
			theLine = "";
			BOOL first = TRUE, Labeled = FALSE;
			int len = 0;
			int count = 0;

			while (nv_it != (*nit).second.end()) 
			{
				CNetListNode& theNode = *nv_it;
				++ nv_it;

				if (theNode.m_label != "" && !Labeled) 
				{
					theLabel = theNode.m_label;
					Labeled = TRUE;
				}
				else
				{
					CString add;
					if (!theNode.m_label.IsEmpty())
					{
					// NOT FOR Protel:	add = theNode.m_label;
					}
					else if (!theNode.m_reference.IsEmpty())
					{
						add.Format("%s-%s", theNode.m_reference, theNode.m_pin );
					}

					if (!add.IsEmpty())
					{
						len += add.GetLength();
						if (len > 127)
						{
							theLine += "\n";
							len = add.GetLength();
							first = TRUE;
						}

						if (first)
						{
							first = FALSE;
						}
						else
						{
							theLine += "\n";
							len ++;
						}


						theLine += add;
						++ count;
						
					}
				}
			}

			if (count > 1) 
			{
				fprintf(theFile,"(\n");
				if (Labeled)
					fprintf(theFile,"%s",(const char *)theLabel);
				else
					fprintf(theFile,"N%06d",Label++);
				fprintf(theFile,"\n%s\n)\n",(const char *)theLine);
			}
		}

		++ nit;
	}


  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
  fclose(theFile);
}


// Create netlist and output as a PCB file (PADS-PCB)
void CNetList::WriteNetListFilePADS( CTinyCadMultiDoc *pDesign, const char *filename )
{
  FILE *theFile = fopen(filename,"w");
  if (theFile == NULL) 
  {
	Message(IDS_CANNOTOPEN);
	return;
  }

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

  // Get the net list
  MakeNet( pDesign );

  fprintf(theFile,"*PADS-PCB*\n");


  fprintf(theFile,"*PART*\n");
  
  // Generate a component for every sheet in this design
  for (int i = 0; i < pDesign->GetNumberOfSheets(); i++)
  {

	  CString last = "";
	  drawingIterator it = pDesign->GetSheet(i)->GetDrawingBegin();
	  while (it != pDesign->GetSheet(i)->GetDrawingEnd()) 
	  {
		CDrawingObject *pointer = *it;

		if (pointer->GetType() == xMethodEx3 && ((CDrawMethod*)pointer)->GetRef()!=last) 
		{
			CDrawMethod *pMethod = static_cast<CDrawMethod *>(pointer);
			CString Ref  = pMethod->GetRef();
			CString Package = "This part has no 'Package' attribute";
			for (int i = 2; i < pMethod->GetFieldCount(); i++)
			{
				if (pMethod->GetFieldName(i).CompareNoCase("package") == 0)
				{
					Package = pMethod->GetField(i);
				}
			}


			// Pad to correct length...
			do
			{
				Ref = Ref + " ";
			}
			while (Ref.GetLength() < 8);
			fprintf(theFile,"%s%s\n", Ref, Package);

			last = ((CDrawMethod *)pointer)->GetRef();
		}
		
		++ it;
	  }
  }


  fprintf(theFile,"\n*NET*\n");
  int Label = 0;

  netCollection::iterator nit = m_nets.begin();

	while (nit != m_nets.end()) 
	{
		nodeVector::iterator nv_it = (*nit).second.begin();

		CString theLine,theLabel;

		if (nv_it != (*nit).second.end()) 
		{
			theLine = "";
			BOOL first = TRUE, Labeled = FALSE;
			int len = 0;
			int count = 0;

			while (nv_it != (*nit).second.end()) 
			{
				CNetListNode& theNode = *nv_it;
				++ nv_it;

				if (theNode.m_label != "" && !Labeled) 
				{
					theLabel = theNode.m_label;
					Labeled = TRUE;
				}
				else
				{
					CString add;
					if (!theNode.m_label.IsEmpty())
					{
					// NOT FOR PADS:	add = theNode.m_label;
					}
					else if (!theNode.m_reference.IsEmpty())
					{
						add.Format("%s.%s", theNode.m_reference, theNode.m_pin );
					}

					if (!add.IsEmpty())
					{
						len += add.GetLength();
						if (len > 127)
						{
							theLine += "\n";
							len = add.GetLength();
							first = TRUE;
						}

						if (first)
						{
							first = FALSE;
						}
						else
						{
							theLine += " ";
							len ++;
						}


						theLine += add;
						++ count;
						
					}
				}
			}

			if (count > 1) 
			{
				fprintf(theFile,"*SIGNAL*  ");
				if (Labeled)
					fprintf(theFile,"%s",(const char *)theLabel);
				else
					fprintf(theFile,"N%06d",Label++);
				fprintf(theFile,"\n%s\n",(const char *)theLine);
			}
		}

		++ nit;
	}

	fprintf(theFile,"*END*\n");


  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
  fclose(theFile);
}


// Create netlist and output as a PCB file
void CNetList::WriteNetListFileTinyCAD( CTinyCadMultiDoc *pDesign, const char *filename )
{
  FILE *theFile = fopen(filename,"w");
  if (theFile == NULL) 
  {
	Message(IDS_CANNOTOPEN);
	return;
  }

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

  // Get the net list
  MakeNet( pDesign );

  fprintf(theFile,NetComment" ====+  Net List for %s  +====\n\n",(const char *)(pDesign->GetPathName()));


  fprintf(theFile,NetComment" ======+ The component list\n\n");

  // Generate a component for every sheet in this design
  for (int i = 0; i < pDesign->GetNumberOfSheets(); i++)
  {
	  CString last = "";
	  drawingIterator it = pDesign->GetSheet(i)->GetDrawingBegin();
	  while (it != pDesign->GetSheet(i)->GetDrawingEnd()) 
	  {
		CDrawingObject *pointer = *it;

		if (pointer->GetType() == xMethodEx3 && ((CDrawMethod*)pointer)->GetRef()!=last) 
		{
			CDrawMethod *pMethod = static_cast<CDrawMethod *>(pointer);
			const char *Name = pMethod->GetField(CDrawMethod::Name);
			const char *Ref  = pMethod->GetRef();
			fprintf(theFile,"COMPONENT '%s' = %s\n",Ref,Name);

			// Now write it it's "other" references
			for (int i = 2; i < pMethod->GetFieldCount(); i++)
			{
				fprintf(theFile,"\tOPTION '%s' = %s\n",pMethod->GetFieldName(i), pMethod->GetField(i) );
			}
			last = ((CDrawMethod *)pointer)->GetRef();
		}
		
		++ it;
	  }
  }

  fprintf(theFile,"\n\n"NetComment" ======+ The net list\n\n");
  int Label = 0;

  netCollection::iterator nit = m_nets.begin();

	while (nit != m_nets.end()) 
	{
		nodeVector::iterator nv_it = (*nit).second.begin();

		CString theLine,theLabel;

		if (nv_it != (*nit).second.end()) 
		{
			theLine = "";
			BOOL first = TRUE,PrintLine=FALSE, Labeled = FALSE;

			while (nv_it != (*nit).second.end()) 
			{
				CNetListNode& theNode = *nv_it;
				++ nv_it;

				if (theNode.m_label != "" && !Labeled) 
				{
					theLabel = theNode.m_label;
					Labeled = TRUE;
				}
				else
				{
					CString add;
					if (!theNode.m_label.IsEmpty())
					{
						add = theNode.m_label;
					}
					else if (!theNode.m_reference.IsEmpty())
					{
						add.Format("(%s,%s)", theNode.m_reference, theNode.m_pin );
					}

					if (!add.IsEmpty())
					{
						if (first)
						{
							first = FALSE;
						}
						else
						{
							theLine += ",";
						}

						theLine += add;
						PrintLine=TRUE;
					}
				}
			}
			if (PrintLine) {
				fprintf(theFile,"NET  ");
				if (Labeled)
					fprintf(theFile,"'%s'",(const char *)theLabel);
				else
					fprintf(theFile,"'N%06d'",Label++);
				fprintf(theFile," =  %s\n",(const char *)theLine);
			}
		}

		++ nit;
	}


  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
  fclose(theFile);
}

// Create netlist and output as a Eagle PCB script
void CNetList::WriteNetListFileEagle( CTinyCadMultiDoc *pDesign, const char *filename )
{
  FILE *theFile = fopen(filename,"w");
  if (theFile == NULL) 
  {
	Message(IDS_CANNOTOPEN);
	return;
  }

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

  // Get the net list
  MakeNet( pDesign );

  int y_pos = 1;
  int x_pos = 1;

  // Generate a component for every sheet in this design
  for (int i = 0; i < pDesign->GetNumberOfSheets(); i++)
  {
	  CString last = "";
	  drawingIterator it = pDesign->GetSheet(i)->GetDrawingBegin();
	  while (it != pDesign->GetSheet(i)->GetDrawingEnd()) 
	  {
		CDrawingObject *pointer = *it;

		if (pointer->GetType() == xMethodEx3 && ((CDrawMethod*)pointer)->GetRef()!=last) 
		{
			CDrawMethod *pMethod = static_cast<CDrawMethod *>(pointer);
			CString Ref  = pMethod->GetRef();
			CString Package = "This part has no 'Package' attribute";
			for (int i = 2; i < pMethod->GetFieldCount(); i++)
			{
				if (pMethod->GetFieldName(i).CompareNoCase("package") == 0)
				{
					Package = pMethod->GetField(i);
				}
			}


			fprintf(theFile,"ADD '%s' %s R0 (0.%d 0.%d);\n", Ref, Package, x_pos, y_pos );

			++ x_pos;
			if (x_pos == 10)
			{
				++ y_pos;
				x_pos = 1;
			}

			last = ((CDrawMethod *)pointer)->GetRef();
		}
		
		++ it;
	  }
  }

  fprintf(theFile,"\n\n");

  netCollection::iterator nit = m_nets.begin();
  int Label = 0;

	while (nit != m_nets.end()) 
	{
		nodeVector::iterator nv_it = (*nit).second.begin();

		CString theLine,theLabel;

		if (nv_it != (*nit).second.end()) 
		{
			theLine = "";
			BOOL first = TRUE,PrintLine=FALSE, Labeled = FALSE;

			while (nv_it != (*nit).second.end()) 
			{
				CNetListNode& theNode = *nv_it;
				++ nv_it;

				if (theNode.m_label != "" && !Labeled) 
				{
					theLabel = theNode.m_label;
					Labeled = TRUE;
				}
				else
				{
					CString add;
					if (!theNode.m_label.IsEmpty())
					{
					}
					else if (!theNode.m_reference.IsEmpty())
					{
						add.Format("   %s %s\n", theNode.m_reference, theNode.m_pin );
					}

					if (!add.IsEmpty())
					{
						theLine += add;
						PrintLine=TRUE;
					}
				}
			}
			if (PrintLine) {
				fprintf(theFile,"SIGNAL ");
				if (Labeled)
					fprintf(theFile,"%s",(const char *)theLabel);
				else
					fprintf(theFile,"N%06d",Label++);
				fprintf(theFile,"\n%s   ;\n",(const char *)theLine);
			}
		}

		++ nit;
	}




  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
  fclose(theFile);
}


// Create netlist and output as a SPICE file
void CNetList::WriteSpiceFile( CTinyCadMultiDoc *pDesign, const char *filename )
{

	FILE *theFile = fopen(filename,"w");
	if (theFile == NULL) 
	{
		Message(IDS_CANNOTOPEN);
		return;
	}


	//
	// TODO!!
	//
	// We should bring up a dialogue here so that the user
	// can enter the parameter for the spice simulation.
	//
	// We would then write those parameters to the top of the
	// spice file.
	//
	//


	// Set the Busy icon
	SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

	// Create the net list
	MakeNet( pDesign );

	
	// Now we have the net list we must convert it into a file suitable for
	// spice...
	//
	// The netlist represents a single vector per net, however, spice requires
	// a vector per symbol.  We have to rotate the array, so that we have one
	// entry in our vector for each of the symbols in our drawing....
	//
	
	typedef std::map<CString,CNetListSymbol>	symbolCollection;
	typedef std::map<int,CString>				labelCollection;
	symbolCollection symbols;
	labelCollection labels;


	netCollection::iterator nit = m_nets.begin();

	while (nit != m_nets.end()) 
	{
		nodeVector::iterator nv_it = (*nit).second.begin();

		while (nv_it != (*nit).second.end()) 
		{
			CNetListNode& theNode = *nv_it;
			++ nv_it;

			// Is this node a symbol?
			if (!theNode.m_reference.IsEmpty())
			{
				// Yes, so update the pin allocations in the symbol map...
				CNetListSymbol &symbol = symbols[ theNode.m_reference ];

				symbol.m_pins[ theNode.m_pin ] = theNode.m_NetList;
				symbol.m_pMethod = theNode.m_pMethod;
			}

			// Is this node a label?
			if (!theNode.m_label.IsEmpty())
			{
				// Yes, so update the label collection
				labels[ theNode.m_NetList ] = theNode.m_label;
			}
		}

		++ nit;
	}


	//
	// TODO!!
	//
	// We should scan the symbols array and extact any file imports
	// That we need from the fields of the symbols...
	//



	// We now have the netlist in the form we require it, so
	// let us now output the symbols in the correct SPICE format...
	//
	symbolCollection::iterator sit = symbols.begin();
	while (sit != symbols.end())
	{
		CString reference = (*sit).first;
		CNetListSymbol &symbol = (*sit).second;
		CDrawMethod* pMethod = symbol.m_pMethod;

		// Here is the data we are going to extract from
		// the symbol's fields...
		CString spice = "NO_MODEL";
		CString name = "NO_VALUE";

		// Search this symbols fields and extract the SPICE field...
		//
		for (int i = 0; i < pMethod->GetFieldCount(); i++)
		{
			if (pMethod->GetFieldName(i).CompareNoCase("spice") == 0)
			{
				spice = pMethod->GetField(i);
			}

			if (pMethod->GetFieldName(i).CompareNoCase("name") == 0)
			{
				name = pMethod->GetField(i);
			}
		}

		// Ok, we have a symbol, it's reference and it's fields
		// so we are in a position to generate the SPICE output line

		// First output the SPICE model
		fprintf(theFile,"%s ", spice);

		// Now output the pins in the order of their name
		//
		// TODO!!!
		// 
		// NOTE: this isn't very helpful, the "spice" parameter really ought
		//       to have some kind of parameters in it e.g. R $(1) $(2) $(NAME) etc....
		//
		//		 Each of the parameters refers to either a pin name or a field name.
		//       We would then replace each of the parameters with the values from the
		//       fields or the netlist number.
		//
		//       Also we haven't dealt with power connections here either.  To determine
		//       If a connection is power connection, use the labels maps.
		//
		pinCollection::iterator pin_it = symbol.m_pins.begin();
		while (pin_it != symbol.m_pins.end())
		{
			CString pin_name = (*pin_it).first;
			int netlist = (*pin_it).second;

			fprintf(theFile, "%d ", netlist );
			
			++ pin_it;
		}

		// Now output the value of this element (the NAME value)
		fprintf( theFile, "%s\n", name );

		++ sit;
	}



	SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
	fclose(theFile);
}

