/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#ifndef __OBJECT_H__
#define __OBJECT_H__

// This class is defined in library.h
class CDesignFileSymbol;
class CTinyCadDoc;

#include "DrawingObject.h"
#include "ArcPoint.h"
#include "DragUtils.h"


// Some sizes of parts...

#define CONNECT_SIZE 4



class CDrawRectOutline : public CDrawingObject
{
protected:

	// The constructor
	CDrawRectOutline( CTinyCadDoc *pDesign )
	: CDrawingObject( pDesign )
	{
	}


	// Redraw this object (including the handles)
	virtual void Display( BOOL erase = TRUE );
	virtual void PaintHandles( Context &dc );

	// Move fields of this object about
	virtual int IsInsideField(CPoint p);
	virtual void MoveField(int w, CPoint r);
	virtual int SetCursorEdit( CPoint p );
};

////// These are the schematic objects //////

class CDrawLine : public CDrawingObject 
{
protected:
	ObjType xtype;

	WORD Style;
	BOOL has_placed;
	BOOL is_stuck;
	BOOL is_junction;
	BOOL re_edit;

	void ToAngle();
	CPoint GetStickyPoint( CPoint q );

	CDragUtils	m_drag_utils_a;
	CDragUtils	m_drag_utils_b;


public:

	CDrawLine& operator=( const CDrawLine &o )
	{
		a = o.a;
		b = o.b;
		xtype = o.xtype;
		Style = o.Style;
		has_placed = o.has_placed;
		is_stuck = o.is_stuck;
		is_junction = o.is_junction;
		re_edit = o.re_edit;
		return *this;
	}


	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag(ObjType t);

	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual ObjType GetType();
	virtual void Paint(Context &dc,paint_options options);
	virtual void Display( BOOL erase = TRUE );
	virtual void LButtonDown(CPoint,CPoint);
	virtual void DblLButtonDown(CPoint,CPoint);
	virtual BOOL CanEdit() { return TRUE; }
	virtual int  GetContextMenu();
	virtual void ContextMenu( CPoint p, UINT id );


	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual CDrawingObject* Store();
	virtual void BeginEdit(BOOL re_edit);
 	virtual void EndEdit();
	virtual void NewOptions();
	virtual CString GetName() const;
	virtual void TagResources();
	virtual int getMenuID();

	virtual void LButtonUp(CPoint);// The user has released the left hand button
	virtual BOOL CanControlAspect();				// If the user holds down ctrl, can should we fix the aspect ratio?

	// Move fields of this object about
	virtual int IsInsideField(CPoint p);
	virtual void MoveField(int w, CPoint r);
	virtual int SetCursorEdit( CPoint p );
	virtual void PaintHandles( Context &dc );


	// This is used for the construction of this object
	CDrawLine(CTinyCadDoc *pDesign, ObjType NewType = xLineEx2);
};




class CDrawPolygon : public CDrawingObject 
{
protected:
	ObjType xtype;

	BOOL m_re_edit;

	WORD Style;
	WORD Fill;

	void ToAngle();
	BOOL IsInsidePolygon( CPoint a );
	BOOL IsInsideLine( int,int,int,int );
	void FindNearestSegment( CPoint a, int &line, int &handle );

	arcpointCollection	m_handles;

	pointCollection		m_points;

	void AddPolyBezier( pointCollection &cp, CArcPoint p1, CArcPoint p2 );
	void CalcBoundingRect();
	void FlatternPath();

public:
	virtual double DistanceFromPoint( CPoint p );

	virtual void Load(CStream &);
	
	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();


	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual ObjType GetType();
	virtual void Paint(Context &dc,paint_options options);
	virtual void Display( BOOL erase = TRUE );
	virtual void LButtonDown(CPoint,CPoint);
	virtual void DblLButtonDown(CPoint,CPoint);
	virtual void Rotate(CPoint p,int dir);

	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual CDrawingObject* Store();
	virtual void BeginEdit(BOOL re_edit);
 	virtual void EndEdit();
	virtual void NewOptions();
	virtual CString GetName() const;
	virtual void TagResources();
	virtual int getMenuID();
	virtual BOOL CanEdit() { return TRUE; }

	virtual void LButtonUp(CPoint);// The user has released the left hand button

	// Move fields of this object about
	virtual int IsInsideField(CPoint p);
	virtual void MoveField(int w, CPoint r);
	virtual int SetCursorEdit( CPoint p );
	virtual void PaintHandles( Context &dc );

	// For the context menu
	virtual void FinishDrawing( CPoint p );
	virtual void ContextMenu( CPoint p, UINT id );
	virtual int  GetContextMenu();

	// This is used for the construction of this object
	CDrawPolygon(CTinyCadDoc *pDesign, ObjType NewType = xLineEx2);
};


class CDrawJunction : public CDrawingObject 
{
protected:
	BOOL is_stuck;
public:
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void Display( BOOL erase = TRUE );
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();
	
	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();


	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual ObjType GetType();		// Get this object's type
	virtual CString GetName() const;
	virtual int getMenuID() { return IDM_TOOLJUNC; }

	// This is used for the construction of this object
	CDrawJunction(CTinyCadDoc *pDesign);
};

class CDrawBusSlash : public CDrawingObject {
	BYTE theDir;
public:
	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual ObjType GetType();		// Get this object's type
	virtual CString GetName() const;
	virtual int getMenuID();
	virtual void Display( BOOL erase = TRUE );


	// These are used for the construction of this object
	CDrawBusSlash(CTinyCadDoc *pDesign,int NewDir = 0);
};

class CDrawNoConnect : public CDrawingObject 
{
protected:
	BOOL is_stuck;
public:
	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void Display( BOOL erase = TRUE );
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();


	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual ObjType GetType();		// Get this object's type
	virtual CString GetName() const;
	virtual int getMenuID() { return IDM_TOOLCONNECT; }

	// This is used for the construction of this object
	CDrawNoConnect(CTinyCadDoc *pDesign);
	virtual ~CDrawNoConnect() { }
};


class CEditDlgTextEdit;

class CDrawText : public CDrawRectOutline
{

	friend CEditDlgTextEdit;

	// The font style
	hFONT	FontStyle;

	// The font colour
	COLORREF	FontColour;

	CString str;
	BYTE dir;
	ObjType xtype;

	BOOL is_stuck;
	
	int	original_width;
	int original_box_width;
	int target_box_width;

	void SetScalingWidths();

public:
	virtual void TagResources();
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag(ObjType t);

	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual ObjType GetType();			// Get this object's type
	virtual void Rotate(CPoint,int);
	virtual BOOL CanEdit();
	virtual void Display( BOOL erase = TRUE );


	CString Find(const char *);	// Does this string match this text?
	virtual CString GetName() const;
	CString GetValue() { return str; }
	void NewFont(LOGFONT *);	// Change the font
	virtual void NewOptions();

	virtual int getMenuID();
	virtual void MoveField(int w, CPoint r);

	// These are used for the construction of this object
	CDrawText(CTinyCadDoc *pDesign,ObjType NewType = xLabelEx2);
	virtual ~CDrawText() { }
};



class CEditDlgPowerEdit;

class CDrawPower : public CDrawingObject {

	friend CEditDlgPowerEdit;

	CPoint	TextPos; // The position of the text item
	CPoint	TextEnd;

	BOOL is_stuck;
	BOOL is_junction;

	CString str;
	BYTE dir;	// Direction this item is facing (0=up,1=down,2=left,3=right)
	BYTE which;	// Which type of power item it is

	int DoRotate(int olddir,int newdir);

public:
	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void TagResources();
	virtual void Display( BOOL erase = TRUE );
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual void Rotate(CPoint,int);
	virtual BOOL CanEdit();
	virtual ObjType GetType();		// Get this object's type
	virtual CString GetName() const;
	CString GetValue() { return str; }
	CString Find(const char *);

	virtual int getMenuID() { return IDM_TOOLPOWER; }

	// These are used for the construction of this object
	CDrawPower( CTinyCadDoc *pDesign );
	virtual ~CDrawPower() { }
};


class CEditDlgPinEdit;
class CDrawMethod;

class CDrawPin : public CDrawingObject {

	friend CEditDlgPinEdit;	// The friendly dialog box to edit this pin

	CString str;	// The pin name
	CString number;	// The pin number
	BYTE show;	// Show name or number?
	BYTE dir;	// Direction this item is facing (0=up,1=down,2=left,3=right)
	BYTE which;	// Which type of power item it is
	BYTE elec;	// Which electrical properties this pin has
	BYTE part;	// Which subpart the item is in
	int	 number_pos;	// The relative position of the pin's number
	WORD length;		// The overall length of the pin

	void DetermineLayout( CPoint &pa,CPoint &pb,CPoint &pc,CPoint &pd,CPoint &pta,CPoint &ptb, int &dr, int rotmir );
	void DetermineSize();
	int DoRotate(int olddir,int newdir);

public:
	int IsInvisible();	// Is this pin currently visible?
			   		// Is this pin a power pin?
	int IsPower() { return which == 4; }

	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void TagResources();
	virtual void Display( BOOL erase = TRUE );
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void OldLoad(CStream &);
	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual BOOL CanEdit();
	virtual void Rotate(CPoint,int);
	virtual ObjType GetType();
	virtual CString GetName() const;
	CString GetNumber()  { return number; }
	CString GetPinName() { return str; }
	CString Find(const char *);
	int     GetPart()    { return part; }
	int		GetElec()    { return elec; }

	virtual int getMenuID() { return IDM_LIBPIN; }

	// Get the actual position of this pin given the parent method
	CPoint GetPosition(CDrawMethod *parent);

	void SetPart(int NewPart) { part=NewPart; }

	CDrawPin(CTinyCadDoc *pDesign);				// The constructoe
	virtual ~CDrawPin() { }
};

////// An imported advanced metafile //////

class CDrawMetaFile : public CDrawRectOutline
{
	hMETAFILE m_metafile;

public:
	virtual ObjType GetType();
	virtual void TagResources();
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual CString GetName() const;
	void setMetaFile( HENHMETAFILE data );

	// This is used for the construction of this object
	CDrawMetaFile(CTinyCadDoc *pDesign);
	virtual ~CDrawMetaFile();

	virtual int getMenuID() { return IDM_TOOLSQUARE; }
};

////// These are the anotation objects //////

class CDrawSquare : public CDrawRectOutline
{
	WORD Style;
	WORD Fill;
	BOOL m_re_edit;
	ObjType m_type;

	double EllipseDistanceFromPoint( CPoint p, BOOL &IsInside  );

public:
	virtual double DistanceFromPoint( CPoint p );
	BOOL PointInEllipse( CPoint p );
	virtual ObjType GetType();
	virtual void TagResources();
	virtual void Paint(Context &,paint_options);
	virtual CDrawingObject* Store();

	void OldLoad(CStream &,int);
	void OldLoad2(CStream &);
	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag( BOOL isSquare );

	virtual void NewOptions();
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual CString GetName() const;
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual BOOL CanEdit() { return TRUE; }
	virtual void LButtonUp(CPoint p);
	virtual void LButtonDown(CPoint,CPoint);
	bool IsSquare() { return m_type == xSquareEx3; }


	// This is used for the construction of this object
	CDrawSquare(CTinyCadDoc *pDesign, ObjType type);

	virtual int getMenuID() { return IsSquare() ? IDM_TOOLSQUARE : IDM_TOOLCIRCLE; }
};


class CDrawError : public CDrawingObject {
	int		ErrorNumber;
public:
	virtual ObjType GetType();
	virtual void Paint(Context &,paint_options);
	virtual CString GetName() const;
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual BOOL CanEdit();
	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual void Display( BOOL erase );
	int GetErrorNumber() { return ErrorNumber; }
	virtual CDrawingObject* Store();

	// This is used for the construction of this object
	CDrawError(CTinyCadDoc *pDesign,CPoint NewA, int q );

};



// The Tag class

class CDrawTag : public CDrawingObject 
{
	CString 	TagName;

public:
	ObjType 	GetType() 		{ return xTag; }
	virtual BOOL IsInside(int,int,int,int) { return FALSE; }

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();


	CString		GetTagName() 	{ return TagName; }
	virtual void		SetTagName(const char *NewTagName) { TagName = NewTagName; }

	// This is used for the construction of this object
	CDrawTag(CTinyCadDoc *pDesign, CPoint NewA,const char *NewTagName)
		: CDrawingObject(pDesign) 
	{ 
		a = NewA; 
		TagName = NewTagName; 
	};

	CDrawTag(CTinyCadDoc *pDesign)
		: CDrawingObject(pDesign)
	{ 
		a=CPoint(0,0); 
		TagName = ""; 
	}
	virtual ~CDrawTag() { }
};





class CDrawRuler : public CDrawingObject 
{
	BOOL	horiz;
	CRect	text_rect;

public:
	virtual double DistanceFromPoint( CPoint p );

	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual BOOL IsInside(int left,int right,int top,int bottom);
	virtual ObjType GetType();
	virtual void Paint(Context &dc,paint_options options);
	virtual void Display( BOOL erase = TRUE );
	virtual void LButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual CDrawingObject* Store();
	virtual CString GetName() const;
	virtual int getMenuID();
	virtual int SetCursorEdit( CPoint p );

	virtual BOOL IsConstruction();

	// This is used for the construction of this object
	CDrawRuler( CTinyCadDoc *pDesign, BOOL new_horiz );
	virtual ~CDrawRuler();
};





////// These are the block objects //////

class CDrawBlockMove : public CDrawingObject {
	int placed;
	CPoint OldPos;
public:
	virtual void Paint(Context &,paint_options);
	CDrawBlockMove(CTinyCadDoc *);
	virtual void EndEdit();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);

	virtual int SetCursor( CPoint p );
	virtual int getMenuID() { return IDM_EDITMOVE; }
};



class CDrawBlockRotate : public CDrawingObject {
	int placed;

public:
	virtual void Paint(Context &,paint_options);
	CDrawBlockRotate(CTinyCadDoc *);
	virtual void EndEdit();
	virtual void BeginEdit(BOOL re_edit);
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void ChangeDir(int);

	virtual int SetCursor( CPoint p );
	virtual int getMenuID() { return IDM_EDITROTATE; }
};


class CDrawBlockDup : public CDrawingObject {
	int placed;

public:
	virtual void Paint(Context &,paint_options);
	CDrawBlockDup(CTinyCadDoc *);
	virtual void EndEdit();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);

	virtual int SetCursor( CPoint p );
	virtual int getMenuID() { return IDM_EDITDUP; }
};



class CDrawBlockDrag : public CDrawingObject 
{
	int placed;
	CPoint			m_OldPos;
	CDragUtils		m_drag_utils;

public:
	virtual void Paint(Context &,paint_options);
	CDrawBlockDrag(CTinyCadDoc *pDesign);
	virtual void EndEdit();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual int SetCursor( CPoint p );
	virtual int getMenuID() { return IDM_EDITDRAG; }
};



class CDrawBlockImport : public CDrawingObject {

public:
	CDrawBlockImport(CTinyCadDoc *pDesign);
	virtual void EndEdit();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void Import( BOOL file_load );

	virtual int getMenuID() { return IDM_FILEIMPORT; }
};



////// These are the single object editing objects //////


class CDrawEditItem : public CDrawingObject {
	BOOL	InMove;
	BOOL	InSelectByDrag;
	int		EditMethodText;
	CPoint  LastPos;

	CDragUtils	m_drag_utils;

	CDrawingObject* GetClosestObject( CPoint p );

	void ClickSelection( CPoint p, CPoint s );

public:
	CDrawEditItem( CTinyCadDoc *pDesign );
	virtual void NewOptions();
	virtual void EndEdit();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);
	virtual void LButtonUp(CPoint);
	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void Paint(Context &,paint_options);
	virtual void ReleaseSelection();
	virtual void EndSelection();
	virtual void RButtonUp(CPoint,CPoint);
	virtual ObjType GetType() { return xEditItem; }
	virtual int SetCursor( CPoint p );
	virtual void BeginEdit( BOOL re_edit );
	virtual void Display( BOOL erase = TRUE );
	virtual void ContextMenu( CPoint p, UINT id );

	virtual int getMenuID() { return IDM_EDITEDIT; }
};



////// This is the method CDrawingObject for use with library symbols //////

class CEditDlgMethodEdit;

class CDrawMethod : public CDrawingObject {

	friend CEditDlgMethodEdit;		// Both these objects aid the edit of this object

	// Where this symbol data comes from
	hSYMBOL			m_Symbol;
	CString	SymName;		// The name of this symbol

	BYTE part;				// Which part in the package this is
	BYTE rotate;			// rotation and mirror
	BOOL can_scale;

	struct CField
	{
		SymbolFieldType	m_type;
		CString			m_description;
		CString			m_value;
		BOOL			m_show;
		CPoint			m_position;
	};

	typedef std::vector<CField>	fieldCollection;
	fieldCollection	m_fields;

	double scaling_x;			// The scaling of this symbol
	double scaling_y;

	void NewRotation();		// Calculate new co-ords after a rotation

	void GetSymbolByName();


	int DoRotate(int olddir,int newdir);

public:
	void TagResources();

	enum FieldPos { Ref=0, Name, Other };

	CPoint GetTr();
	CDesignFileSymbol *GetSymbolData();
	BOOL ExtractSymbol( CPoint &tr, drawingCollection &method );
	virtual BOOL IsInside(int left,int right,int top,int bottom);

	virtual double DistanceFromPoint( CPoint p );
	virtual int IsInsideField(CPoint);		// Is a point pointing at one of our fields?
	virtual void MoveField(int,CPoint);		// Move a field
	virtual int SetCursorEdit( CPoint p );

	void Display( BOOL erase = TRUE );
	virtual void Paint(Context &,paint_options);
	virtual void PaintHandles( Context &dc );
	virtual CDrawingObject* Store();

	void OldLoad(CStream &);
	void OldLoad2(CStream &);
	void OldLoad3(CStream &);
	virtual void Load(CStream &);

	virtual void SaveXML( CXMLWriter &xml );
	virtual void LoadXML( CXMLReader &xml );
	static const char* GetXMLTag();

	virtual void Move(CPoint, CPoint no_snap_p);
	virtual void LButtonDown(CPoint,CPoint);
	virtual void BeginEdit(BOOL re_edit);
	virtual void EndEdit();
	virtual BOOL CanEdit();
	virtual void Rotate(CPoint,int);
	BOOL IsNoSymbol();

	virtual ObjType GetType();		// Get this object's type
	virtual CString GetName() const;

	CString GetRef() const { return m_fields[Ref].m_value; }
	CString GetField(int);		// Get the string value of a field
	int		GetFieldCount() { return m_fields.size(); }
	CString	GetFieldName(int);
	CString Find(const char *);
	CPoint  GetFieldPos(int q) { return CPoint(a.x+m_fields[q].m_position.x,a.y+m_fields[q].m_position.y); }
	BOOL	HasRef() { return strchr(GetRef(),'?') == NULL; }

	int  GetRefVal();
	int  GetSubPart() { return part; }
	int  GetRotate() { return rotate; }
	virtual void SetRefVal(int value);
	void SetRef(const char *NewRef) { m_fields[Ref].m_value = NewRef; }
	void SetPart(int NewPart) { part = NewPart; }
	hSYMBOL GetSymbolID() { return m_Symbol; }

	// Add/remove the next references
	void AddReference( int min_ref );
	void RemoveReference();

	CDrawMethod(CTinyCadDoc *pDesign,hSYMBOL,int new_rotation);		// The constructors
	CDrawMethod(CTinyCadDoc *pDesign);
	virtual ~CDrawMethod() 
	{
	}

	virtual int getMenuID() { return IDM_TOOLGET; }
};


////// These are non-displayable objects! //////

// The zoom in/out object
class CDrawCentre : public CDrawingObject {
public:
	CDrawCentre( CTinyCadDoc *pDesign )
		: CDrawingObject( pDesign )
	{
	}

	virtual void Display( BOOL erase );
	virtual ObjType GetType();
	virtual void LButtonDown(CPoint,CPoint);		// Zoom In
	virtual BOOL RButtonDown(CPoint,CPoint);				// Zoom Out

	virtual int SetCursor( CPoint p ) { return 9; }
	virtual int getMenuID() { return IDM_VIEWCENTRE; }
};


// The zoom in/out object
class CDrawRefPainter : public CDrawingObject 
{
private:
	int m_starting_point;

public:
	CDrawRefPainter( CTinyCadDoc *pDesign, int starting_point )
		: CDrawingObject( pDesign )
	{
		m_starting_point = starting_point;
	}

	virtual void Display( BOOL erase );
	virtual ObjType GetType();
	virtual void LButtonDown(CPoint,CPoint);
	virtual BOOL RButtonDown(CPoint,CPoint);				// Zoom Out
	virtual void RButtonUp(CPoint,CPoint);// The user has released the right hand button
	virtual int SetCursor( CPoint p ) { return 10; }
	virtual int getMenuID() { return -1; }
};


#endif
