/*
 * Project:		TinyCAD program for schematic capture
 *				http://tinycad.sourceforge.net
 * Copyright:	 1994-2005 Matt Pyne
 * License:		Lesser GNU Public License 2.1 (LGPL)
 *				http://www.opensource.org/licenses/lgpl-license.html
 */

#include "stdafx.h"
#include "TinyCadView.h"
#include "Option.h"
#include "Registry.h"
#include "Colour.h"

//-------------------------------------------------------------------------
//-- The option catorgories
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
COption::COption()
{
	PinLength = 20;
	PinNumberPos = 0;
}
//-------------------------------------------------------------------------
void COption::Init( CTinyCadDoc *pDesign )
{
	// Keep a track of the design
	m_pDesign = pDesign;

	// Init the font settings for normal text, pin text and ruler text
	theFontList = ListOfFonts( ListOfFonts::TEXT_HEIGHT, false );
	theFontList.Add(new ListOfFonts(ListOfFonts::PIN_HEIGHT,false ));
	theFontList.Add(new ListOfFonts(ListOfFonts::RULE_HEIGHT, true));

	// Init the line settings for normal and dashed lines
	theStyleList = ListOfStyles( PS_SOLID, 1, cBLACK );
	theStyleList.Add( new ListOfStyles(PS_DASH, 1, cBLACK) );

	// Init no fill and black fill styles
	theFillStyleList = ListOfFillStyles( -1, cBLACK );
	theFillStyleList.Add(new ListOfFillStyles(0,cBLACK));

	CurrentFont = 0;
	CurrentStyle = 0;
	CurrentFillStyle = 0;

	// Attempt to load the defaults for the grid
	GridShow			= CRegistry::GetBool("ShowGrid", false);
	Units				= CRegistry::GetInt("Units", 0);
	ScrollBar			= CRegistry::GetInt("ScrollBar", 20);
	PinLength			= CRegistry::GetInt("PinLength", 30);
	PinNumberPos		= CRegistry::GetInt("PinNumberPos", 0);
	AutoDrag			= CRegistry::GetBool("AutoDrag", true );
	AutoJunc			= CRegistry::GetBool("AutoJunc", true);
	AutoSnap			= CRegistry::GetBool("AutoSnap", true);
	AutoSnapRange		= CRegistry::GetInt("AutoDragRange", 15);

	// Create the no symbol object
	CDrawSquare *Sq = new CDrawSquare(pDesign,xSquareEx3);
	Sq->a=CPoint(0,0);
	Sq->b=CPoint(-40,-40);
	CPoint tr = CPoint(-40,-40);
	drawingCollection method;
	method.push_back( Sq );
	CDesignFileSymbol *NoSymbol = new CDesignFileSymbol( pDesign, method, tr );
	NoSymbol->name = "No Symbol Defined";
	NoSymbol->reference="?";
	NoSymbol->ppp = 1;  
	theSymbolList = new ListOfSymbols( NoSymbol );
}

// The destructor
COption::~COption()
{
	delete theSymbolList;
}

void COption::UnTag()
{
	theFontList.UnTag();
	theStyleList.UnTag();
	theFillStyleList.UnTag();
	theMetaFileList.UnTag();
	theSymbolList->UnTag();
}

////// Get the user defined colours //////

//-------------------------------------------------------------------------
COLORREF COption::GetColour( int u )
{
	return m_oColors.Get( u );
}
//-------------------------------------------------------------------------
void COption::SetColour( int n, COLORREF cr )
{
	m_oColors.Set( n, cr );
}
//-------------------------------------------------------------------------
// Set the default colours (ignoring the registry)
void COption::ResetColours()
{
	m_oColors.Reset();
}
//-------------------------------------------------------------------------
CUserColor& COption::GetUserColor()
{
	return m_oColors;
}
//-------------------------------------------------------------------------

////// Some default settings for dialogs ////// 

void COption::SetPinLength( int pl )
{
	PinLength = pl;
	CRegistry::Set("PinLength", PinLength );
}

void COption::SetPinNumberPos( int pl )
{
	PinNumberPos = pl;
	CRegistry::Set("PinNumberPos", PinNumberPos );
}

////// Save the current settings along with the design //////

void COption::SaveOptions( CStream& theArchive )
{
	theArchive << (BYTE) xOptions;
	theArchive << CurrentFont;
	theArchive << CurrentStyle;
	theArchive << GridShow;
	theArchive << Units;
	theArchive << ScrollBar;
}

void COption::SaveOptionsXML(CXMLWriter &xml)
{
	xml.addTag( "OPTIONS" );

	xml.addTag( "FONT", CurrentFont );
	xml.addTag( "STYLE", CurrentStyle );
	xml.addTag( "GRID", GridShow );
	xml.addTag( "UNITS", Units );
	xml.addTag( "SCROLLBAR", ScrollBar );
	m_oColors.Write( xml );

	xml.closeTag();
}

void COption::LoadOptionsXML( CXMLReader &xml, BOOL keep )
{
	CString name;
	xml.intoTag();

	ResetColours();

	while (xml.nextTag( name ))
	{
		if (name == "FONT")
		{
			xml.getChildData( CurrentFont );
		}
		else if (name == "STYLE")
		{
			xml.getChildData( CurrentStyle );
		}
		else if (name == "GRID")
		{
			xml.getChildData( GridShow );
		}
		else if (name == "UNITS")
		{
			xml.getChildData( Units );
		}
		else if (name == "SCROLLBAR")
		{
			xml.getChildData( ScrollBar );
		}
		else
		{
			m_oColors.Read( xml, name );
		}
	}	

	xml.outofTag();
}

void COption::LoadOptions( CStream &theArchive, BOOL keep )
{
	if (keep)
	{
		theArchive >> CurrentFont;
		theArchive >> CurrentStyle;
		theArchive >> GridShow;
		theArchive >> Units;
		theArchive >> ScrollBar;
		CurrentFont = GetNewFontNumber( CurrentFont );
		CurrentStyle = GetNewStyleNumber( CurrentStyle );
	}
	else
	{
		hFONT	dCurrentFont;
		hSTYLE	dCurrentStyle;
		BOOL	dGridShow;
		LONG	dUnits;
		LONG	dScrollBar;
		theArchive >> dCurrentFont;
		theArchive >> dCurrentStyle;
		theArchive >> dGridShow;
		theArchive >> dUnits;
		theArchive >> dScrollBar;
	}
}

////// The edit options //////

//-------------------------------------------------------------------------
//-- Change the size of a font and create a new entry
hFONT COption::ChangeFontSize( hFONT n, int delta_height, int width )
{
  LOGFONT Font = *GetFont( n );
  Font.lfHeight += delta_height;

  if (width != 0)
  {
	Font.lfWidth = width;
  }

  return AddFont( &Font );
}
//-------------------------------------------------------------------------
int COption::GetFontWidth( hFONT n )
{
	LOGFONT* Font = GetFont( n );

  // If we have an aspect ratio that is not 1.0 then we must
  // first find the "average" width of the font.  We can only
  // do this by first selecting the font with an aspect ratio
  // of 1.0 and measuring it.

  if (Font->lfWidth == 0)
  {
		CClientDC dc( AfxGetMainWnd() );
	CFont f;
	f.CreateFontIndirect( Font );

	TEXTMETRIC metrics;
		CFont* old_font = dc.SelectObject( &f );
	dc.GetTextMetrics( &metrics );
	dc.SelectObject( old_font );
	Font->lfWidth = metrics.tmAveCharWidth;
  }

  return Font->lfWidth;
}
//-------------------------------------------------------------------------
void COption::ChooseFont(CDC &PrinterDC, HWND Window)
{
	CHOOSEFONT cf;
	LOGFONT aLogFont = *(GetFont(GetCurrentFont()));

	cf.lStructSize = sizeof(CHOOSEFONT);
    cf.hwndOwner = Window;
    cf.hDC = PrinterDC.m_hDC;
    cf.lpLogFont = &aLogFont;
    cf.iPointSize = 10;
    cf.Flags = CF_FORCEFONTEXIST | CF_INITTOLOGFONTSTRUCT | CF_BOTH;
    cf.rgbColors = cBLACK;
    cf.lCustData = 0;
    cf.lpfnHook = NULL;
    cf.lpTemplateName = NULL;
    cf.hInstance = NULL;
    cf.lpszStyle = NULL;
    cf.nFontType = SCREEN_FONTTYPE;
    cf.nSizeMin = 5;
    cf.nSizeMax = 128;

	if (::ChooseFont(&cf))
	{
		CurrentFont = AddFont(&aLogFont);
	}
}

////// The symbol operators //////

void COption::ResetSymbols()
{
	ListOfSymbols *pSymbol = theSymbolList;

	while (pSymbol != NULL)
	{
		pSymbol->ResetSymbol();
		pSymbol = static_cast<ListOfSymbols*>(pSymbol->next);
	}
}

// Convert a CPoint to a Unit CString
CString COption::PointToDisplay(CPoint a,BOOL horiz)
{
 	CString r;

  	float Scale = GetUnits()==0 ? (float)PIXELSPERMM : ((float)PIXELSPERMM * 25.4f);
	float sx = a.x / Scale;
	float sy = a.y / Scale;
	if (horiz)
	{
  		sprintf(r.GetBuffer(20),"%.1f",sy,GetUnits()==0 ? "mm" : "\"");
	}
	else
	{
		sprintf(r.GetBuffer(20),"%.1f",sx,GetUnits()==0 ? "mm" : "\"");
	}
	r.ReleaseBuffer();

	return r;
}


// Convert a CPoint to a Unit CString
CString COption::PointToUnit(CPoint a)
{
 	CString r;

  	float Scale = GetUnits()==0 ? (float)PIXELSPERMM : ((float)PIXELSPERMM * 25.4f);
	float sx = a.x / Scale;
	float sy = a.y / Scale;
  	sprintf(r.GetBuffer(20),"%.1f,%.1f%s",sx,sy,GetUnits()==0 ? "mm" : "\"");
	r.ReleaseBuffer();

	return r;
}
//-------------------------------------------------------------------------
void COption::ResetMerge()
{
	theFontList.ResetMerge();
	theStyleList.ResetMerge();
	theFillStyleList.ResetMerge();
	theMetaFileList.ResetMerge();
	theSymbolList->ResetMerge();
}
//-------------------------------------------------------------------------
hFONT COption::GetCurrentFont() const
{
	return CurrentFont;
}
//-------------------------------------------------------------------------
hFONT COption::AddFont( LOGFONT* lf, hFONT MergeNumber)
{
	return theFontList.Add( new ListOfFonts(*lf), MergeNumber );
}
//-------------------------------------------------------------------------

