/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __OPTION_H__
#define __OPTION_H__



#include "Symbol.h"
#include "DocResource.h"
#include "ListOfFonts.h"
#include "ListOfStyles.h"
#include "ListOfFillStyles.h"
#include "ListOfMetaFiles.h"
#include "ListOfSymbols.h"
#include "UserColor.h"

//=========================================================================
class COption
{
	CTinyCadDoc*		m_pDesign;

	ListOfFonts			theFontList;
	ListOfStyles		theStyleList;
	ListOfFillStyles	theFillStyleList;
	ListOfMetaFiles		theMetaFileList;
	ListOfSymbols*		theSymbolList;

	// The current settings
	hFONT	CurrentFont;						// The default font number
	hSTYLE	CurrentStyle;						// The default line style
	hFILL	CurrentFillStyle;					// The default line style

	// The current grid settings
	Bool8	GridShow;							// Should we show the grid?
	LONG	Units;								// The units to use
	LONG	ScrollBar;							// The scroll bar increments

	// Some default settings for dialogs
	int		PinLength;							// The default pin length
	int		PinNumberPos;						// The default pin number position

	// The current auto-snap settings
	BOOL	AutoSnap;
	BOOL	AutoDrag;
	BOOL	AutoJunc;
	int		AutoSnapRange;

	CUserColor m_oColors;

public:
	COption();
	~COption();

	void Init( CTinyCadDoc *pDesign );

	// Untag all of our resources
	void UnTag();

	////// The font operators ////

	//-- Add a font to the list and return it's new index number
	hFONT AddFont(LOGFONT *lf,hFONT MergeNumber = -1);

	//-- Load the font table from a file
	void LoadFonts(CStream &theFile) { theFontList.Load(m_pDesign,theFile); }

	// Convert an old font number into a new font number
	hFONT GetNewFontNumber(hFONT n) { return theFontList.ConvertOld(n); }

	//-- Get a font from the number
	LOGFONT *GetFont(hFONT n)  { return ((ListOfFonts *)theFontList.Get(n))->GetFontPtr(); }
	hFONT ChangeFontSize(hFONT n, int delta_height, int width );

	//-- Specify this font is in use
	void TagFont(hFONT n) { theFontList.Tag(n); }

	//-- Bring up the user change font box
	void ChooseFont(CDC &,HWND = NULL);	
	hFONT  	GetCurrentFont() const;
	void SetCurrentFont(hFONT n) { CurrentFont = n; }
	int GetFontWidth( hFONT n );

	////// The Line Style operators //////

	// Add a line style to the list and return it's new index number
	hSTYLE AddStyle(LineStyle *ls,hFONT MergeNumber = -1) { return theStyleList.Add(new ListOfStyles(*ls), MergeNumber); }
	// Load the line style table from a file
	void LoadStyles(CStream &theFile) { theStyleList.Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hSTYLE GetNewStyleNumber(hSTYLE n) { return theStyleList.ConvertOld(n); }
	// Get a line style from the number
	LineStyle *GetStyle(hSTYLE n)  { return ((ListOfStyles *)theStyleList.Get(n))->GetLineStylePtr(); }
	// Specify this line style is in use
	void TagStyle(hSTYLE n) { theStyleList.Tag(n); }
	hSTYLE  GetCurrentStyle() { return CurrentStyle; }
	void SetCurrentStyle(hSTYLE n) { CurrentStyle = n; }


	////// The Fill Style operators //////

	// Add a fill style to the list and return it's new index number
	hFILL AddFillStyle(FillStyle *ls,hFONT MergeNumber = -1) { return theFillStyleList.Add(new ListOfFillStyles(*ls), MergeNumber); }
	// Load the fill style table from a file
	void LoadFillStyles(CStream &theFile) { theFillStyleList.Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hFILL GetNewFillStyleNumber(hFILL n) { return theFillStyleList.ConvertOld(n); }
	// Get a fill style from the number
	FillStyle *GetFillStyle(hSTYLE n)  { return ((ListOfFillStyles *)theFillStyleList.Get(n))->GetFillStylePtr(); }
	// Specify this fill style is in use
	void TagFillStyle(hSTYLE n) { theFillStyleList.Tag(n); }
	hFILL  GetCurrentFillStyle() { return CurrentFillStyle; }
	void SetCurrentFillStyle(hFILL n) { CurrentFillStyle = n; }

	////// The Meta File operators //////

	// Add a line style to the list and return it's new index number
	hMETAFILE AddMetaFile(HENHMETAFILE s,hMETAFILE MergeNumber = -1) { return theMetaFileList.Add(new ListOfMetaFiles(s), MergeNumber); }
	// Load the line style table from a file
	void LoadMetaFiles(CStream &theFile) { theMetaFileList.Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hMETAFILE GetNewMetaFileNumber(hMETAFILE n) { return theMetaFileList.ConvertOld(n); }
	// Get a line style from the number
	HENHMETAFILE GetMetaFile(hMETAFILE n)  { return ((ListOfMetaFiles *)theMetaFileList.Get(n))->GetMetaFile(); }
	// Specify this line style is in use
	void TagMetaFile(hSTYLE n) { theMetaFileList.Tag(n); }

	////// The Symbols operators //////

	// Add a line style to the list and return it's new index number
	hSYMBOL AddSymbol(CDesignFileSymbol *s,hSYMBOL MergeNumber = -1) { return theSymbolList->Add(new ListOfSymbols(s), MergeNumber); }
	// Load the line style table from a file
	void LoadSymbols(CStream &theFile) { theSymbolList->Load(m_pDesign,theFile); }
	// Convert an old line style number into a new font number
	hSYMBOL GetNewSymbolNumber(hSYMBOL n) { return theSymbolList->ConvertOld(n); }
	// Get a line style from the number
	CDesignFileSymbol *GetSymbol(hSYMBOL n)  { return ((ListOfSymbols *)theSymbolList->Get(n))->GetSymbol(); }
	// Specify this line style is in use
	void TagSymbol(hSYMBOL n) { theSymbolList->Tag(n); }
	// Reload the symbols from their libraries
	void ResetSymbols();

	////// Save the current settings along with the design //////
	void SaveOptions( CStream &theArchive );
	void LoadOptions( CStream &theArchive, BOOL keep );

	////// Get the user defined colours //////
	void ResetColours();
	COLORREF GetColour( int u );
	void SetColour( int u, COLORREF c );
	CUserColor& GetUserColor();

	////// The XML operations //////

	void SaveFontsXML(CXMLWriter &xml) { theFontList.SaveXML(m_pDesign,"FONT", xml); }
	void SaveStylesXML(CXMLWriter &xml) { theStyleList.SaveXML(m_pDesign,"STYLE", xml); }
	void SaveFillStylesXML(CXMLWriter &xml) { theFillStyleList.SaveXML(m_pDesign,"FILL", xml); }
	void SaveMetaFilesXML(CXMLWriter &xml) { theMetaFileList.SaveXML(m_pDesign,"META", xml); }
	void SaveSymbolsXML(CXMLWriter &xml) { theSymbolList->SaveXML(m_pDesign,"SYMBOLDEF", xml); }

	void SaveOptionsXML(CXMLWriter &xml);

	void LoadFontXML(CXMLReader &xml) { theFontList.LoadItemXML(m_pDesign,xml); }
	void LoadStyleXML(CXMLReader &xml) { theStyleList.LoadItemXML(m_pDesign,xml); }
	void LoadFillStyleXML(CXMLReader &xml) { theFillStyleList.LoadItemXML(m_pDesign,xml); }
	void LoadMetaFileXML(CXMLReader &xml) { theMetaFileList.LoadItemXML(m_pDesign,xml); }
	void LoadSymbolXML(CXMLReader &xml) { theSymbolList->LoadItemXML(m_pDesign,xml); }

	void LoadOptionsXML(CXMLReader &xml, BOOL keep);

	void ResetMerge();

	////// The grid operators //////

	bool ShowGrid() { return GridShow == TRUE ; }		// The size of the current grid
	void SetGridShow( bool s ) { GridShow = s; }
	int	 GetUnits() { return Units; }
	void SetUnits( int u ) { Units = u; }
	int  GetScrollBar() { return ScrollBar; }
	void SetScrollBar( int s ) { ScrollBar = s; }
	CString PointToUnit(CPoint);
	CString PointToDisplay(CPoint,BOOL horiz);

////// The auto-wire settins //////
	bool GetAutoSnap() { return AutoSnap == TRUE; }
	bool GetAutoDrag() { return AutoDrag == TRUE; }
	bool GetAutoJunc() { return AutoJunc == TRUE; }
	int	 GetAutoSnapRange() { return AutoSnapRange; }

	void SetAutoSnap(bool r) { AutoSnap = r; }
	void SetAutoDrag(bool r) { AutoDrag = r; }
	void SetAutoJunc(bool r) { AutoJunc = r; }
	void SetAutoSnapRange( int r ) { AutoSnapRange = r; }

	////// Some default settings for dialogs ////// 
	int		GetPinLength() { return PinLength; }
	void	SetPinLength( int pl );

	int		GetPinNumberPos() { return PinNumberPos; }
	void	SetPinNumberPos( int pl );
};
//=========================================================================

#endif
