/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include <stdlib.h>
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "LineUtils.h"

// The Power Class //

#define POWER_SIZE	14



void CDrawPower::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(fPIN);
}


int CDrawPower::DoRotate(int olddir,int newdir)
{
  //     New rotation=> 2  3  4  
  //				    Current dir ..\/..
  int table[] = {	2, 3, 0, 		// 0 (Up)
			3, 2, 1,		// 1 (Down)
			1, 0, 3,		// 2 (Left)
			0, 1, 2,		// 3 (Right)
		};

  return table[(newdir-2) + olddir*3];
}

const char* CDrawPower::GetXMLTag()
{
	return "POWER";
}

// Load and save to an XML file
void CDrawPower::SaveXML( CXMLWriter &xml )
{
	xml.addTag(GetXMLTag());

	xml.addAttribute( "pos", CDPoint(a) );
	xml.addAttribute( "which", which );
	xml.addAttribute( "direction", dir );
	xml.addChildData( str );

	xml.closeTag();
}

void CDrawPower::LoadXML( CXMLReader &xml )
{
	xml.getAttribute( "pos", a );
	xml.getAttribute( "which", which );
	xml.getAttribute( "direction", dir );
	xml.getChildData( str );

	b = a;
	TextPos = a;
}


void CDrawPower::Load(CStream &archive)
{
 	a = ReadPoint(archive);
  archive >> which >> dir >> str;
  b=a;
  m_segment=0;
}

double CDrawPower::DistanceFromPoint( CPoint p )
{
	CPoint pa;

  // Find out which way round the power object goes
  switch (dir) {
	case 0:		// Top
		pa=CPoint(a.x,a.y-POWER_SIZE);
		break;
	case 1:		// Bottom
		pa=CPoint(a.x,a.y+POWER_SIZE);
		break;
	case 2:		// Left
		pa=CPoint(a.x-POWER_SIZE,a.y);
		break;
	case 3:		// Right
		pa=CPoint(a.x+POWER_SIZE,a.y);
		break;
  }

  CLineUtils l(a,pa);
  CPoint d;
  return l.DistanceFromPoint(p, d );
}



BOOL CDrawPower::IsInside(int left,int right,int top,int bottom)
{
	CPoint pa;

  // Find out which way round the power object goes
  switch (dir) {
	case 0:		// Top
		pa=CPoint(a.x,a.y-POWER_SIZE);
		break;
	case 1:		// Bottom
		pa=CPoint(a.x,a.y+POWER_SIZE);
		break;
	case 2:		// Left
		pa=CPoint(a.x-POWER_SIZE,a.y);
		break;
	case 3:		// Right
		pa=CPoint(a.x+POWER_SIZE,a.y);
		break;
  }

  CLineUtils l(a,pa);

  return !( (TextPos.x<left && TextEnd.x<=left) || (TextPos.x>right  && TextEnd.x>=right)
         || (TextPos.y<top  && TextEnd.y<=top)  || (TextPos.y>bottom && TextEnd.y>=bottom) )
         || l.IsInside(left,right,top,bottom);
}

// Look for a seach string in the object
CString CDrawPower::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = str;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? str : "";
}


// Rotate this object about a point
void CDrawPower::Rotate(CPoint p,int ndir)
{			 
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  dir = DoRotate(dir,ndir);
}


BOOL CDrawPower::CanEdit()
{
  return TRUE;
}

ObjType CDrawPower::GetType()
{
  return xPower;
}

CString CDrawPower::GetName() const
{
  return "Power";
}


CDrawPower::CDrawPower(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
  m_segment=1;
  a=CPoint(0,0);
  b=a;
  str="";
  dir=1;
  which=0;
  is_stuck = FALSE;
  is_junction = FALSE;
}

void CDrawPower::BeginEdit(BOOL re_edit)
{
  g_EditToolBar.m_PowerEdit.Open(m_pDesign,this);
}

void CDrawPower::Move(CPoint p, CPoint no_snap_p)
{
  Display();
  p = m_pDesign->GetStickyPoint(p, TRUE, TRUE, is_stuck, is_junction);
  a=p;
  b=a;
  m_segment=0;
  Display();
}

void CDrawPower::LButtonDown(CPoint p, CPoint)
{
  // New undo level for each placement...
  m_pDesign->BeginNewChangeSet();

  Display();
  p = m_pDesign->GetStickyPoint(p, TRUE, TRUE, is_stuck, is_junction);
  a=p;
  Store();
  Display();	// Write to screen
}


void CDrawPower::Display( BOOL erase )
{
	// Invalidate the symbol
	CRect r( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r, erase, POWER_SIZE * 2 + 2 );

	r.left = TextPos.x;
	r.top = TextPos.y;
	r.right = TextEnd.x;
	r.bottom = TextEnd.y;
	m_pDesign->InvalidateRect( r, erase, 5 );

 if (is_stuck)
 {
	 r.left = b.x - HIGHLIGHT_SIZE;
	 r.right = b.x + HIGHLIGHT_SIZE;
	 r.top = b.y - HIGHLIGHT_SIZE;
	 r.bottom = b.y + HIGHLIGHT_SIZE;
	 m_pDesign->InvalidateRect( r, erase, 3 );
 }

}


// Display the power item on the screen!
void CDrawPower::Paint(Context &dc,paint_options options)
{
  int spacing;
  CSize size;
  CPoint pa,pb,pc,pd,pos=a;

  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(fPIN),2);
  dc.SetROP2(R2_COPYPEN);
  dc.SetTextColor(m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::POWER));

  size=dc.GetTextExtent(str);
  if (which!=0)
	spacing=POWER_SIZE*2+POWER_SIZE/4;
  else
	spacing=POWER_SIZE+POWER_SIZE/4;

  // Find out which way round the power object goes
  switch (dir) {
	case 0:		// Top
		pa=CPoint(pos.x,pos.y-POWER_SIZE);
		pb=CPoint(pos.x-POWER_SIZE/2,pos.y-POWER_SIZE);
		pc=CPoint(pos.x+POWER_SIZE/2+1,pos.y-POWER_SIZE);
		pd=CPoint(pos.x,pos.y-POWER_SIZE*2);
		TextPos = CPoint(pos.x-size.cx/2,pos.y-spacing);
		break;
	case 1:		// Bottom
		pa=CPoint(pos.x,pos.y+POWER_SIZE);
		pb=CPoint(pos.x-POWER_SIZE/2,pos.y+POWER_SIZE);
		pc=CPoint(pos.x+POWER_SIZE/2+1,pos.y+POWER_SIZE);
		pd=CPoint(pos.x,pos.y+POWER_SIZE*2);
		TextPos = CPoint(pos.x-size.cx/2,pos.y+spacing+size.cy);
		break;
	case 2:		// Left
		pa=CPoint(pos.x-POWER_SIZE,pos.y);
		pb=CPoint(pos.x-POWER_SIZE,pos.y-POWER_SIZE/2);
		pc=CPoint(pos.x-POWER_SIZE,pos.y+POWER_SIZE/2+1);
		pd=CPoint(pos.x-POWER_SIZE*2,pos.y);
		TextPos=CPoint(pos.x-size.cx-spacing,pos.y+size.cy/2);
		break;
	case 3:		// Right
		pa=CPoint(pos.x+POWER_SIZE,pos.y);
		pb=CPoint(pos.x+POWER_SIZE,pos.y-POWER_SIZE/2);
		pc=CPoint(pos.x+POWER_SIZE,pos.y+POWER_SIZE/2+1);
		pd=CPoint(pos.x+POWER_SIZE*2,pos.y);
		TextPos=CPoint(pos.x+spacing,pos.y+size.cy/2);
		break;
  }

  dc.TextOut(str,TextPos,options);
  TextEnd = CPoint(TextPos.x + size.cx,TextPos.y - size.cy);

  switch (options)
  {
  case draw_selected:
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cSELECT);
	  break;
  default:
	  dc.SelectPen(PS_SOLID,1,cLINE);
  }

  
  dc.SelectBrush();


  // Draw the main line of the power item
  switch (which) {
	case 0: // Draw the Bar
  		dc.MoveTo(pb);
  		dc.LineTo(pc);
		break;
	case 1: // Draw the Circle
		dc.Ellipse(CRect(pb.x-(pa.x-pd.x),pb.y-(pa.y-pd.y),pc.x,pc.y));
		break;
	case 2: // Draw the Wave
		// The PolyBezier function is not implemented in MFC!
		// So this hack has to be used....
		{
		  POINT pts[4];
		  pts[0]=CPoint(pb.x+(pd.x-pa.x)/2,pb.y+(pd.y-pa.y)/2);
		  pts[1]=CPoint(pd.x+(pb.x-pa.x)/2,pd.y+(pb.y-pa.y)/2);
		  pts[2]=CPoint(pa.x+(pc.x-pa.x)/2,pa.y+(pc.y-pa.y)/2);
		  pts[3]=CPoint(pc.x-(pa.x-pd.x)/2,pc.y-(pa.y-pd.y)/2);
		  dc.PolyBezier(pts,4);
		}
		break;
	case 3: // Draw the Arrow
		dc.MoveTo(pb);
  		dc.LineTo(pc);
		dc.MoveTo(pb);
		dc.LineTo(pd);
		dc.LineTo(pc);
		break;
	case 4: // Draw the Earth
  		dc.MoveTo(pb);
  		dc.LineTo(pc);
		dc.MoveTo(CPoint(pb.x-(pb.x-pd.x)/2,pb.y-(pb.y-pd.y)/2));
		dc.LineTo(CPoint(pc.x-(pc.x-pd.x)/2,pc.y-(pc.y-pd.y)/2));
		dc.MoveTo(CPoint(pd.x-(pb.x-pa.x)/4,pd.y-(pb.y-pa.y)/4));
		dc.LineTo(CPoint(pd.x+(pb.x-pa.x)/4,pd.y+(pb.y-pa.y)/4));
		break;
  }



  // Draw the bits that all power items have in comman
  dc.MoveTo(pos);
  dc.LineTo(pa);


  if (is_stuck)
  {
	  // Draw a nice circle to show the stickness...
	  dc.SelectBrush();
	  dc.SelectPen(PS_SOLID,5,cBOLD);
	  dc.SetROP2(R2_COPYPEN);
	  dc.Ellipse(CRect(b.x+HIGHLIGHT_SIZE,b.y+HIGHLIGHT_SIZE,b.x-HIGHLIGHT_SIZE,b.y-HIGHLIGHT_SIZE));

	  // Do we need a junction
	  if (is_junction)
	  {
	    int js=JUNCTION_SIZE;
		CPoint br,tl;
		br=CPoint(b.x+js+1,b.y+js+1);
		tl=CPoint(b.x-js,b.y-js);

		dc.SetROP2(R2_COPYPEN);

		dc.SelectPen(PS_SOLID,1,m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::JUNCTION));
		dc.SelectBrush(m_pDesign->GetOptions()->GetUserColor().Get( CUserColor::JUNCTION));
  	    dc.Ellipse(CRect(tl.x,tl.y,br.x,br.y));
	  }
  }

}


void CDrawPower::EndEdit()
{
	RButtonDown(CPoint(0,0), CPoint(0,0));
	g_EditToolBar.m_PowerEdit.Close();
}

// Store the line in the drawing
CDrawingObject* CDrawPower::Store()
{
  CDrawPower *NewObject;

  NewObject = new CDrawPower(m_pDesign);
  *NewObject = *this;
  NewObject->is_junction = FALSE;
  NewObject->is_stuck = FALSE;

  m_pDesign->Add(NewObject);
  if (is_junction)
  {
	  m_pDesign->AddNewJunction( a );
  }

  return NewObject;
}


