/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"

#define theWnd ((CMainWindow)dc.theWindow)


////// The Line Class //////


CDrawLine::CDrawLine(CTinyCadDoc *pDesign, ObjType NewType)
: CDrawingObject( pDesign )
{
	segment=1;
	xtype=NewType;
	Undo_Level=0;
	Delete_Level=-1;

	switch (xtype) {
		case xBus:
			Style = fBUS;
			break;
		case xLineEx2:
			Style = m_pDesign->GetOptions()->GetCurrentStyle();
			break;
		case xDash:
			Style = fDASH;
			break;
		case xWire:
			Style = fWIRE;
			break;
		default:
			Style = fLINE;
	}	
}


// Save the line to a file
void CDrawLine::Save(CStream& archive )
{
  WritePoint(archive,a);
  WritePoint(archive,b);
	if (xtype == xLineEx2)
  		archive << Style;
  	
}


void CDrawLine::TagResources()
{
   m_pDesign->GetOptions()->TagStyle(Style);
}


// Load the line from a file
void CDrawLine::Load(CStream& archive )
{
 	a = ReadPoint(archive);
   	b = ReadPoint(archive);

	switch (xtype) {
		case xDash:
		case xLine:
			// Premote this type
			xtype = xLineEx2;
			break;
		case xLineEx:
			// Load the line thickness
			BYTE LineThickness;
			archive >> LineThickness;
			LineStyle l;
			l.Colour = cBLACK;
			l.Style = PS_SOLID;
			l.Thickness = LineThickness;
			Style = m_pDesign->GetOptions()->AddStyle(&l);
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			xtype = xLineEx2;
			break;
		case xLineEx2:
			archive >> Style;
			Style = m_pDesign->GetOptions()->GetNewStyleNumber(Style);
			break;
	}	
}


void CDrawLine::ToAngle()
{
  int FromX,FromY,FromC;

  switch (g_EditToolBar.m_LineEdit.GetAngle()) {
	case LINEBOX_90:
		if (a.x==b.x)
			g_EditToolBar.m_LineEdit.mode=2;
		if (a.y==b.y)
			g_EditToolBar.m_LineEdit.mode=1;
		if (g_EditToolBar.m_LineEdit.mode==0)
			g_EditToolBar.m_LineEdit.mode=1;
		break;
	case LINEBOX_45:
		g_EditToolBar.m_LineEdit.mode=0;
		// Which 45degree slope is it closest to?
		FromX=abs(a.x-b.x);
		FromY=abs(a.y-b.y);
		FromC=abs(FromX-FromY);
		// Is it closest to the X axis?
		if (FromX<FromY && FromX<FromC)
			b.x=a.x;
		// Is it closest to the Y axis?
		else if (FromY<FromC)
			b.y=a.y;
		// Must be closest to the 45 degree line
		else
			b.x=a.x+(b.x>a.x ? FromY : -FromY);
		break;
	default:
		g_EditToolBar.m_LineEdit.mode=0;
  }
}


BOOL CDrawLine::End()
{
  BOOL r = !segment;

  Display();
  segment=1;
  a = b;

  return r;
}

void CDrawLine::NewOptions()
{
  if (xtype == xLineEx2) {
  	Display();
	Style = m_pDesign->GetOptions()->GetCurrentStyle();
	Display();
  }
}


CString CDrawLine::GetName()
{

  switch (xtype) {
	case xWire:
		return "Wire";
	case xBus:
		return "Bus";
	default:
		return "Line";
  }
}

int CDrawLine::getMenuID() 
{ 
  switch (xtype) {
	case xWire:
		return IDM_TOOLWIRE;
	case xBus:
		return IDM_TOOLBUS;
	default:
		return IDM_TOOLLINE;
  }
}


void CDrawLine::Place(CPoint p, CPoint)
{
  if (segment) {
	a=p;
	b=p;
	segment=!segment;
  } else {
	b=p;
	ToAngle();
	// Remove from screen
	Display();
	CPoint c=b;

	switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: b=CPoint(b.x,a.y);
			break;
		case 2: b=CPoint(a.x,b.y);
			break;
	}
	Store();
	Display();	// Write to screen

	a=b;
	b=c;
	segment=0;
	ToAngle();
  }
  Display();
}


void CDrawLine::Move(CPoint p, CPoint no_snap_p)
{
  if (!segment) {
	Display();
	b=p;
	ToAngle();
	Display();
  }
}


ObjType CDrawLine::GetType()
{
  return xtype;
}

void CDrawLine::Display( BOOL erase )
{
	LineStyle *theStyle = m_pDesign->GetOptions()->GetStyle(Style);
	int width = max(2,theStyle->Thickness) * 2;


  CRect r;
  r.top = a.y;
  r.left = a.x;
  r.bottom = a.y;
  r.right = a.x;

  if (!segment) 
  {
	  switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: 
			r.bottom = a.y;
			r.right = b.x;
			m_pDesign->InvalidateRect( r, erase, width );
			break;
		case 2:
			r.bottom = b.y;
			r.right = a.x;
			m_pDesign->InvalidateRect( r, erase, width );
			break;
	  }
  }

 r.top = r.bottom;
 r.left = r.right;
 r.bottom = b.y;
 r.right = b.x;
 m_pDesign->InvalidateRect( r, erase, width );
 
}


class CLineRectTracker : public CRectTracker
{
public:

	CLineRectTracker( CRect r )
		: CRectTracker( r, CRectTracker::resizeOutside   )
	{
	}

	virtual UINT GetHandleMask( ) const
	{
		return 5;
	}
};

void CDrawLine::DrawHandles( Context &dc )
{
	// Put some handles around this object
	CRect r(a.x,a.y,b.x,b.y);

	CLineRectTracker tracker( dc.theTransform.Scale(r) );
	tracker.Draw( dc.dc );
}


// Move fields of this object about
int CDrawLine::IsInsideField(CPoint p)
{
	CLineRectTracker	tracker( CRect(a.x,a.y,b.x,b.y) );
	int r = tracker.HitTest( p );

	if (r == CLineRectTracker::hitMiddle)
	{
		r = CLineRectTracker::hitNothing;
	}

	return r;
}

void CDrawLine::MoveField(int w, CPoint r)
{
	Display();

	CRect rect( a.x, a.y, b.x, b.y );

	switch (w)
	{
	case CLineRectTracker::hitTopLeft:
		rect.left += r.x;
		rect.top += r.y;
		break;
	case CLineRectTracker::hitTopRight:
		rect.right += r.x;
		rect.top += r.y;
		break;
	case CLineRectTracker::hitBottomRight:
		rect.right += r.x;
		rect.bottom += r.y;
		break;
	case CLineRectTracker::hitBottomLeft:
		rect.left += r.x;
		rect.bottom += r.y;
		break;	
	case CLineRectTracker::hitTop:
		rect.top += r.y;
		break;		
	case CLineRectTracker::hitRight:
		rect.right += r.x;
		break;		
	case CLineRectTracker::hitBottom:
		rect.bottom += r.y;
		break;		
	case CLineRectTracker::hitLeft:
		rect.left += r.x;
		break;		
	}

	a.x = rect.left;
	a.y = rect.top;
	b.x = rect.right;
	b.y = rect.bottom;

	Display();
}

int CDrawLine::SetCursorEdit( CPoint p )
{
	CRect r(a.x,a.y,b.x,b.y);
	r.NormalizeRect();
	CLineRectTracker	tracker( r );
	int s = tracker.HitTest( p );

	if (s == -1 || s == CRectTracker::hitMiddle)
	{
		s = CDrawingObject::SetCursorEdit( p );
	}

	return s;
}



// Display the line on the screen!
void CDrawLine::Paint(Context &dc,paint_options options)
{
  dc.SelectPen(m_pDesign->GetOptions()->GetStyle(Style), options);
 
  dc.SetROP2(R2_COPYPEN);

  dc.MoveTo(a);

  if (!segment) {
	  switch (g_EditToolBar.m_LineEdit.mode) {
		case 1: dc.LineTo(CPoint(b.x,a.y));
			break;
		case 2:	dc.LineTo(CPoint(a.x,b.y));
			break;
	  }
  }
  dc.LineTo(b);

  if (options == draw_handles)
  {
	DrawHandles( dc );
  }
}


// Store the line in the drawing
void CDrawLine::Store()
{
  CDrawLine *NewObject;

  NewObject = new CDrawLine(m_pDesign);
  // Copy the details
  segment=1;
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  CDrawLine *NewRepeat = new CDrawLine(m_pDesign);
  CPoint r=m_pDesign->GetRepeatDir();
  *NewRepeat = *this;
 
  NewRepeat->a = CPoint(a.x+r.x,a.y+r.y);
  NewRepeat->b = CPoint(b.x+r.x,b.y+r.y);

  m_pDesign->SetRepeat(NewRepeat);
}


void CDrawLine::BeginEdit()
{
  segment=1;
  g_EditToolBar.m_LineEdit.Open(m_pDesign,this);
  g_EditToolBar.m_LineEdit.mode=0;

  if (xtype == xLineEx2) {
	m_pDesign->GetOptions()->SetCurrentStyle(Style);
  }
}

void CDrawLine::EndEdit()
{
  Display();
  g_EditToolBar.m_LineEdit.Close();
}


// Work out whether a line is inside or passes through an area:-
//
//                    xxx1     xxxx     xx1x   
//            
// x1xx               1001  |  1000  |  1010
//                    ------+--------+------
// xxxx               0001  |  0000  |  0010
//                    ------+--------+------
// 1xxx               0101  |  0100  |  0110
//
inline int WhichArea(CPoint p,int left,int right,int top,int bottom)
{
  int r=0;

  if (p.x<left)
	r |=1;
  else if (p.x>right)
	r |= 2;

  if (p.y>bottom)
	r |= 4;
  else if (p.y<top)
	r |= 8;

  return r;
}


int CDrawLine::rIsInside(CPoint p1,CPoint p2,int left,int right,int top,int bottom)
{
  {
	int aa = WhichArea(p1,left,right,top,bottom);
	int ab = WhichArea(p2,left,right,top,bottom);

	// Is it a trivial case?
	if (aa==0 || ab==0)
		return 1;
	if ((aa & ab)!=0)
		return 0;
  }

  // Cannot be accepted or rejected, so divide in two and try again!
  {
	CPoint midpoint;
	midpoint.x=(p1.x+p2.x)>>1;
	midpoint.y=(p1.y+p2.y)>>1;
	return (midpoint != p2 && rIsInside(p1,midpoint,left,right,top,bottom)) || (midpoint != p1 && rIsInside(midpoint,p2,left,right,top,bottom));
  }
}

int CDrawLine::IsInside(int left,int right,int top,int bottom)
{
  return rIsInside(a,b,left,right,top,bottom);
}




////// The Junction Class ////

#define JUNCTION_SIZE 3

CDrawJunction::CDrawJunction(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
	a=b=CPoint(0,0);
	segment=1;
	Undo_Level=0;
	Delete_Level=-1;
}


void CDrawJunction::Load(CStream &archive)
{
  a = ReadPoint(archive);
  b=a;
  segment=0;
}


void CDrawJunction::Save(CStream &archive)
{
  WritePoint(archive,a);
}


int CDrawJunction::IsInside(int left,int right,int top,int bottom)
{
  return (a.x>=left && a.x<=right && a.y>=top && a.y<=bottom);
}


ObjType CDrawJunction::GetType()
{
  return xJunction;
}


CString CDrawJunction::GetName()
{
  return "Junction";
}

void CDrawJunction::Move(CPoint p, CPoint no_snap_p)
{
  Display();
  a=p;
  b=a;
  segment=0;
  Display();
}

void CDrawJunction::Place(CPoint p, CPoint)
{
  Display();
  a=p;
  b=p;
  Store();
  Display();	// Write to screen
}

void CDrawJunction::Display( BOOL erase )
{
	// Invalidate the symbol
	CRect r( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r, erase, JUNCTION_SIZE + 1 );
}


void CDrawJunction::Paint(Context &dc,paint_options options)
{
  int js=JUNCTION_SIZE;
  CPoint br,tl;
  br=CPoint(a.x+js+1,a.y+js+1);
  tl=CPoint(a.x-js,a.y-js);

	dc.SetROP2(R2_COPYPEN);


  switch (options)
  {
  case draw_selected:
  case draw_handles:
	  dc.SelectPen(PS_SOLID,1,cSELECT);
      dc.SelectBrush(cSELECT);
	  break;
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cJUNCTION_CLK);
      dc.SelectBrush(cJUNCTION_CLK);
	  break;
  default:
		dc.SelectPen(PS_SOLID,1,cJUNCTION);
		dc.SelectBrush(cJUNCTION);
  }

  dc.Ellipse(CRect(tl.x,tl.y,br.x,br.y));
}


// Store the junction in the drawing
void CDrawJunction::Store()
{
  CDrawJunction *NewObject;

  NewObject = new CDrawJunction(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  m_pDesign->SetRepeat(NULL);
}




////// The No Connect Class //////

#define CONNECT_SIZE 4

CDrawNoConnect::CDrawNoConnect(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
	a=b=CPoint(0,0);
	segment=1;
	Undo_Level=0;
	Delete_Level=-1;
}


void CDrawNoConnect::Load(CStream &archive)
{
 	a = ReadPoint(archive);
  b=a;
  segment=0;
}


void CDrawNoConnect::Save(CStream &archive)
{
  WritePoint(archive,a);
}


int CDrawNoConnect::IsInside(int left,int right,int top,int bottom)
{
  return (a.x>=left && a.x<=right && a.y>=top && a.y<=bottom);
}


ObjType CDrawNoConnect::GetType()
{
  return xNoConnect;
}

CString CDrawNoConnect::GetName()
{
  return "No Connect";
}


void CDrawNoConnect::Move(CPoint p, CPoint no_snap_p)
{
  Display();
  a=p;
  segment=0;
  Display();
}

void CDrawNoConnect::Place(CPoint p, CPoint)
{
  Display();
  a=p;
  b=p;
  Store();
  Display();	// Write to screen
}

void CDrawNoConnect::Display( BOOL erase )
{
	// Invalidate the symbol
	CRect r( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r, erase, CONNECT_SIZE + 1 );
}



void CDrawNoConnect::Paint(Context &dc,paint_options options)
{
  switch (options)
  {
  case draw_selected:
  case draw_handles:
	  dc.SelectPen(PS_SOLID,1,cSELECT);
	  break;
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cNOCONNECT_CLK);
	  break;
  default:
	  dc.SelectPen(PS_SOLID,1,cNOCONNECT);
  }

  
  

	dc.SetROP2(R2_COPYPEN);

  CPoint LineTo1=CPoint(a.x+CONNECT_SIZE,a.y+CONNECT_SIZE);
  CPoint LineTo2=CPoint(a.x+CONNECT_SIZE,a.y-CONNECT_SIZE);

  dc.MoveTo(CPoint(a.x-CONNECT_SIZE,a.y-CONNECT_SIZE));
  dc.LineTo(CPoint(LineTo1.x+1,LineTo1.y));
  dc.MoveTo(CPoint(a.x-CONNECT_SIZE,a.y+CONNECT_SIZE));
  dc.LineTo(CPoint(LineTo2.x+1,LineTo2.y));

}


// Store the no connect in the drawing
void CDrawNoConnect::Store()
{
  CDrawNoConnect *NewObject;

  NewObject = new CDrawNoConnect(m_pDesign);

  *NewObject = *this;

  m_pDesign->Add(NewObject);

  m_pDesign->SetRepeat(NULL);
}



////// The Text Class //////


void CDrawText::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(FontStyle);
}


BOOL CDrawText::CanEdit()
{
  return TRUE;
}

// Look for a seach string in the object
CString CDrawText::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = str;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? str : "";
}

void CDrawText::NewOptions()
{
  Display();
  FontStyle = m_pDesign->GetOptions()->GetCurrentFont();
  Display();
}

void CDrawText::Save(CStream &archive)
{
  WritePoint(archive,a);
  archive << dir << str << FontStyle;
}


int CDrawText::getMenuID()
{
	switch (xtype) {
		case xLabelEx:
		case xLabel:
			return IDM_TOOLLABEL;
		case xTextEx:
		case xText:
			return IDM_TOOLTEXT;
	}

	return -1;
}

void CDrawText::Load(CStream &archive)
{
  a = ReadPoint(archive);
  archive >> dir >> str;
  switch (xtype) {
  	case xLabelEx:
	case xTextEx:
		// Load the font number in
		archive >> FontStyle;

	  	// Convert the font number over
		FontStyle = m_pDesign->GetOptions()->GetNewFontNumber(FontStyle);
		break;
	case xLabel:
		FontStyle = fTEXT;
		xtype = xLabelEx;
		break;
	case xText:
		FontStyle = fTEXT;
		xtype = xTextEx;
		break;
	}


  	CSize size=m_pDesign->GetTextExtent(str,FontStyle);

  if (dir>=2)
	b=CPoint(a.x+size.cx,a.y-size.cy);
  else
	b=CPoint(a.x-size.cy,a.y-size.cx);

  segment=0;
}


ObjType CDrawText::GetType()
{
  return xtype;
}

CString CDrawText::GetName()
{
  if (xtype==xLabelEx)
	return "Label";
  else
	return "Text";
}


void CDrawText::BeginEdit()
{
  g_EditToolBar.m_TextEdit.Open(m_pDesign,this);

  m_pDesign->GetOptions()->SetCurrentFont(FontStyle);
}


void CDrawText::EndEdit()
{
  End();
  g_EditToolBar.m_TextEdit.Close();
}



void CDrawText::Move(CPoint p, CPoint no_snap_p)
{
  Display();			
  a=p;
  segment=0;
  Display();
}


void CDrawText::MoveField(int w, CPoint r)
{
	Display();

	CRect rect( a.x, a.y, b.x, b.y );

	Display();
	switch (w)
	{
	case CRectTracker::hitTopLeft:
		rect.left += r.x;
		rect.top += r.y;
		r.y = -r.y;
		break;
	case CRectTracker::hitTopRight:
		rect.right += r.x;
		rect.top += r.y;
		r.y = -r.y;
		r.x = -r.x;
		break;
	case CRectTracker::hitBottomRight:
		rect.right += r.x;
		rect.bottom += r.y;
		r.x = -r.x;
		break;
	case CRectTracker::hitBottomLeft:
		rect.left += r.x;
		rect.bottom += r.y;
		break;	
	case CRectTracker::hitTop:
		rect.top += r.y;
		r.y = -r.y;
		break;		
	case CRectTracker::hitRight:
		rect.right += r.x;
		r.x = -r.x;
		break;		
	case CRectTracker::hitBottom:
		rect.bottom += r.y;
		break;		
	case CRectTracker::hitLeft:
		rect.left += r.x;
		break;		
	}

	FontStyle = m_pDesign->GetOptions()->ChangeFontSize(FontStyle, dir < 2 ? r.x : r.y );

	a.x = rect.left;
	a.y = rect.top;
	b.x = rect.right;
	b.y = rect.bottom;

	Display();
}

void CDrawText::Place(CPoint p, CPoint)
{
  Display();
  a=p;
  Store();
  Display();	// Write to screen

  // Now increment the number as necessary
  int hold=str.FindOneOf("0123456789-");
  // Only increment if old value was a number at end of string
  if (hold!=-1) {
	char NewNum[17];
	_itoa(atoi(str.Mid(hold))+m_pDesign->GetNameDir(),NewNum,10);
	str = str.Left(hold)+NewNum;
  }
  g_EditToolBar.m_TextEdit.ReFocus();
  Display();
}


// The constructors

CDrawText::CDrawText(CTinyCadDoc *pDesign,ObjType NewType)
: CDrawRectOutline( pDesign )
{
	xtype=NewType;
	dir=3;
	str="";
	segment=1;
	a=b=CPoint(0,0);
	Undo_Level=0;
	Delete_Level=-1;
    FontStyle = m_pDesign->GetOptions()->GetCurrentFont();
}



// Display the text on the screen!
void CDrawText::Paint(Context &dc,paint_options options)
{
  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(FontStyle),dir);

  CSize size=dc.GetTextExtent(str);
  if (dir>=2)
	b=CPoint(a.x+size.cx,a.y-size.cy);
  else
	b=CPoint(a.x-size.cy,a.y-size.cx);

  dc.SetTextColor(xtype==xLabelEx ? cLABEL : cTEXT);
  dc.TextOut(str,a,options,dir);

  if (options == draw_handles)
  {
	  DrawHandles( dc );
  }
}


// Store the text in the drawing
void CDrawText::Store()
{
  CDrawText *NewObject;

  NewObject = new CDrawText(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  m_pDesign->SetRepeat(NULL);
}


// Rotate this object about a point
void CDrawText::Rotate(CPoint p,int ndir)
{
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  if (ndir!=4) {
	if (dir>=2)
		dir = 0;
	else
		dir = 3;
  }

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  CPoint la = a;

  // Adjust for the rotation
  if (dir>=2) {
	a = CPoint(min(la.x,b.x),max(la.y,b.y));
	b = CPoint(max(la.x,b.x),min(la.y,b.y));
  } else {
	a = CPoint(max(la.x,b.x),max(la.y,b.y));
	b = CPoint(min(la.x,b.x),min(la.y,b.y));
  }
}
