/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "TinyCadView.h"


////// The Symbol Field class //////

// Save this symbol field into an archive
void CSymbolField::Save(CStream &ar)
{
	DWORD ft = field_type;
	ar << field_name << field_value << field_default << field_pos << ft;
}

void CSymbolField::SaveXML(CXMLWriter &xml, bool simple)
{
	DWORD ft = field_type;
	xml.addTag("FIELD");
	xml.addTag("NAME", field_name );
	xml.addTag("DEFAULT", field_default );
	xml.addTag("FT", ft );
	if (!simple)
	{
		xml.addTag("VALUE", field_value );
		xml.addTag("POS", CDPoint(field_pos) );
	}
	xml.closeTag();
}


// Load this symbol field from an archive
void CSymbolField::Load(CStream &ar)
{
	DWORD ft;
	ar >> field_name >> field_value >> field_default >> field_pos >> ft;
	field_type = (SymbolFieldType)ft;
}

void CSymbolField::LoadXML(CXMLReader &xml)
{
	DWORD ft = field_type;
	CString name;

	while (xml.nextTag( name ))
	{
		if (name == "NAME")
		{
			xml.getChildData( field_name );
		}
		else if (name == "VALUE")
		{
			xml.getChildData( field_value );
		}
		else if (name == "DEFAULT")
		{
			xml.getChildData( field_default );
		}
		else if (name == "POS")
		{
			xml.getChildData( field_pos );
		}
		else if (name == "FT")
		{
			xml.getChildData( ft );
			field_type = static_cast<SymbolFieldType>(ft);
		}
	}
}

////// The Symbol Record class //////

// The constructor
CSymbolRecord::CSymbolRecord()
{
	name_type = default_show;
	ref_type = default_show;
    fields_loaded = FALSE;
    NameID = -1;
}


// Save this symbol into an XML file
void CSymbolRecord::SaveXML(CXMLWriter &xml)
{
	xml.addTag( "NAME" ); 
	xml.addAttribute( "type", name_type );
	xml.addChildData( name );
	xml.closeTag();

	xml.addTag( "REF" );
	xml.addAttribute( "type", ref_type );
	xml.addChildData( reference );
	xml.closeTag();

	xml.addTag( "DESCRIPTION", description );
	
	for (int i = 0; i < fields.size(); i++)
	{
		fields[i].SaveXML( xml, true );
	}
}

// Load this symbol from an XML file
void CSymbolRecord::LoadXML( CXMLReader &xml)
{
	int nt = 0;
	int rt = 0;
	CString tag_name;
	while (xml.nextTag( tag_name ))
	{
		if (tag_name == "NAME")
		{
			xml.getChildData( name );
			xml.getAttribute( "type", nt );
			name_type = static_cast<SymbolFieldType>(nt);
		}
		else if (tag_name == "REF")
		{
			xml.getChildData( reference );
			xml.getAttribute( "type", rt );
			ref_type = static_cast<SymbolFieldType>(rt);
		}
		else if (tag_name == "DESCRIPTION")
		{ 
			xml.getChildData( description );
		}
		else if (tag_name == "FIELD")
		{
			CSymbolField f;
			xml.intoTag();			
			f.LoadXML( xml );
			xml.outofTag();
			fields.push_back( f );
		}
	}
}


BOOL CSymbolRecord::LoadSymbol( CTinyCadDoc *pDesign, CStream &theArchive, CPoint &tr, drawingCollection &drawing )
{
  // Move to the right location in filespace
  try
  {
	  // Load in the design
	  if (!pDesign->ReadFile( theArchive, FALSE, drawing ))
	  {
		  return FALSE;
	  }
  }
  catch (...)
  {
	  // Failed to load the symbol
	  return FALSE;
  }


  PostLoad(pDesign, tr, drawing);

  return TRUE;
}

void CSymbolRecord::PostLoad(CTinyCadDoc *pDesign, CPoint &tr, drawingCollection &drawing)
{

  // Find the co-ords of the bounding box of this symbol
  CPoint a=CPoint(0,0);
  CPoint b=drawing.front()->a;

  BOOL has_pins = FALSE;
  CPoint pin = CPoint(0,0);

  drawingIterator it = drawing.begin();
  while (it != drawing.end()) 
  {
	  CDrawingObject *pointer = *it;

	if (pointer->GetType()==xPinEx)
	{
		CDrawPin *thePin = static_cast<CDrawPin*>(pointer);

		if (!has_pins)
		{
			if (!thePin->IsInvisible())
			{
				// Use this for the pin-offset calculation
				pin = pointer->a;
				has_pins = TRUE;
			}
		}
	}


	a.x=max(a.x,max(pointer->a.x,pointer->b.x));
	a.y=max(a.y,max(pointer->a.y,pointer->b.y));

	b.x=min(b.x,min(pointer->a.x,pointer->b.x));
	b.y=min(b.y,min(pointer->a.y,pointer->b.y));
	
	++ it;
  }

  // If this symbol has pins, then snap the pin to the grid,
  // rather than the bounding rect
  CPoint snapa, snapb;
  CPoint pin_offset = CPoint(0,0);

  // Snap the co-ords of the bounding box to the grid
  snapa = pDesign->m_snap.Snap(a);
  snapb = pDesign->m_snap.Snap(b);

  // Now make sure the pins stay on-grid
  if (has_pins)
  {
	  // Snap the co-ords of the pin to the grid
	  pin_offset=pin - pDesign->m_snap.Snap(pin);

	  snapa = snapa + pin_offset;
	  snapb = snapb + pin_offset;
  }

  
  // Make sure snapping is always positive
  if (snapa.x<a.x)
	snapa.x+=pDesign->m_snap.GetGrid();
  if (snapa.y<a.y)
	snapa.y+=pDesign->m_snap.GetGrid();
  if (snapb.x>b.x)
	snapb.x-=pDesign->m_snap.GetGrid();
  if (snapb.y>b.y)
	snapb.y-=pDesign->m_snap.GetGrid();

  // Now translate so that bottom left hand corner is at 0,0
  it = drawing.begin();
  while (it != drawing.end()) 
  {
	  CDrawingObject *pointer = *it;
	pointer->a=CPoint(pointer->a.x-snapa.x,pointer->a.y-snapa.y);
	pointer->b=CPoint(pointer->b.x-snapa.x,pointer->b.y-snapa.y);
	
	++ it;
  }

  // Store top right hand corner
  tr=CPoint(snapb.x-snapa.x ,snapb.y-snapa.y );
}




////// The Symbol class //////

// The constructor
CLibraryStoreNameSet::CLibraryStoreNameSet(CLibraryStore *NewLib)
{
  lib=NewLib;
  FilePos = -1;
}

// The destructor
CLibraryStoreNameSet::~CLibraryStoreNameSet()
{
}

CString CLibraryStoreNameSet::GetLibName()
{
  return lib->m_name;
}



// Save this symbol into an archive
void CLibraryStoreNameSet::Save(CStream &theArchive)
{
	CSymbolRecord &r = GetRecord(0);

	DWORD flags = r.name_type << 3 | r.ref_type;
	DWORD number_of_fields = r.fields.size() | (flags << 16);
	theArchive << r.name << r.description << r.reference << ppp << number_of_fields; 

	for (int i = 0; i < r.fields.size(); i++)
	{
	  r.fields[i].Save( theArchive );
	}
}


// Save this NameSet as an XML export
void CLibraryStoreNameSet::SaveXML( CXMLWriter &xml )
{
	xml.addTag( "PPP", ppp );
	xml.addTag( "ORIENTATION", orientation );

	for (int i =0; i < GetNumRecords(); i++)
	{
		xml.addTag("DETAILS");
		CSymbolRecord &r = GetRecord(i);
		r.SaveXML( xml );
		xml.closeTag();
	}
}

// Save this NameSet as an XML export
void CLibraryStoreNameSet::LoadXML(CTinyCadDoc *pDesign, drawingCollection &drawing, CXMLReader &xml)
{
	int nt = 0;
	int rt = 0;
	CString name;

	while (xml.nextTag( name ))
	{
		if (name == "PPP")
		{
			xml.getChildData( ppp );
		}
		else if (name == "ORIENTATION")
		{
			xml.getChildData( orientation );
		}
		else if (name == "DETAILS")
		{
			CLibraryStoreSymbol r;
			xml.intoTag();
			r.LoadXML( xml );
			xml.outofTag();

			PushBackRecord( r );
		}
		else if (name == "TinyCAD")
		{
			pDesign->ReadFileXML(xml, FALSE, drawing, TRUE );
		}
	}
}


// load this symbol from an archive
void CLibraryStoreNameSet::Load(CStream &theArchive)
{
	Blank();
	CSymbolRecord &r = GetRecord(0);


	DWORD number_of_fields;
	FilePos = 0;
	theArchive >> r.name >> r.description >> r.reference >> ppp >> number_of_fields; 
	DWORD flags = number_of_fields >> 16;
	number_of_fields = number_of_fields & 0xffff;
	for (int i = 0; i < number_of_fields; i++)
	{
	  CSymbolField sf;
	  sf.Load( theArchive );
	  r.fields.push_back( sf );
	}

	r.name_type = static_cast<SymbolFieldType>(flags >> 3);
	r.ref_type = static_cast<SymbolFieldType>(flags & 0x7);
}

// load this symbol from an archive
void CLibraryStoreNameSet::OldLoad3(CStream &theArchive)
{
	CString dummy;
	Blank();
	CSymbolRecord &r = GetRecord(0);

    theArchive >> r.name >> r.description >> r.reference >> ppp >> FilePos >> dummy; 
}


// load this symbol from an archive
void CLibraryStoreNameSet::OldLoad1(CStream &theArchive)
{
	Blank();
	CSymbolRecord &r = GetRecord(0);

	theArchive >> r.name >> r.description >> r.reference >> ppp >> FilePos; 
}


// the old version of load this symbol from an archive
void CLibraryStoreNameSet::OldLoad2(CStream &theArchive)
{
	Blank();
	CSymbolRecord &r = GetRecord(0);

	theArchive >> r.name >> r.reference >> ppp >> FilePos; 
	r.description = "";
}



// Get the object this symbol refers to, this method, may not be in memory
// and so will have to be extracted from the library file.
BOOL CLibraryStoreNameSet::GetMethod( CTinyCadDoc *pDesign, CPoint &tr, drawingCollection &drawing )
{
  // No, so must load it from the methods file
  CStream *stream=lib->GetMethodArchive( this );
  if (!stream)
  {
	  return FALSE;
  }
  
  CSymbolRecord &rec = GetRecord(0);
  BOOL r = rec.LoadSymbol( pDesign, *stream, tr, drawing );

  // Get the default orientation
  orientation = stream->GetOrientation();

  delete stream;
  return r;
}



CDesignFileSymbol *CLibraryStoreNameSet::GetDesignSymbol( CTinyCadDoc *pDesign, int index )
{
	CPoint tr;

	drawingCollection drawing;
	if (!GetMethod( pDesign, tr, drawing ))
	{
		return NULL;
	}

	CDesignFileSymbol *psymbol = new CDesignFileSymbol( pDesign, drawing, tr );
    CSymbolRecord &r = GetRecord(index);

	psymbol->description = r.description;
	psymbol->name = r.name;
	psymbol->ppp = ppp;
	psymbol->reference = r.reference;
	psymbol->name_type = r.name_type;
	psymbol->ref_type = r.ref_type;
	psymbol->fields = r.fields;
	psymbol->FilePos = FilePos;

	return psymbol;
}


void CLibraryStoreNameSet::Blank()
{
	m_records.clear();

	CLibraryStoreSymbol r;
	r.description = "New symbol";
	FilePos = -1;
	r.name = "New symbol";
	ppp = 1;
	r.reference = "U?";
	r.NameID = -1;
	r.m_index = 0;
	r.m_pParent = this;
	orientation = 0;

	m_records.push_back( r );
}


CLibraryStoreSymbol& CLibraryStoreNameSet::GetRecord( int index )
{
	CLibraryStoreSymbol &r = m_records[ index ];

	// Enforce the parent parameters...
	r.m_index = index;
	r.m_pParent = this;

	return r;
}

int	CLibraryStoreNameSet::GetNumRecords()
{
	return m_records.size();
}

void CLibraryStoreNameSet::SetRecords( recordCollection& records )
{
	m_records = records;
}

void CLibraryStoreNameSet::PushBackRecord( CLibraryStoreSymbol &r )
{
	r.m_index = m_records.size();
	r.m_pParent = this;
	m_records.push_back( r );
}


///// The symbol as contained by the library set /////


CLibraryStoreSymbol::CLibraryStoreSymbol()
{
	m_index = 0;
	m_pParent = NULL;
}

CDesignFileSymbol *CLibraryStoreSymbol::GetDesignSymbol( CTinyCadDoc *pDesign )
{
	return m_pParent->GetDesignSymbol( pDesign, m_index );
}




///// The symbol as contained by the design /////

CDesignFileSymbol::CDesignFileSymbol(CTinyCadDoc *pDesign, drawingCollection &method, CPoint tr )
{
	m_pDesign = pDesign;
	m_method = method;
	m_Tr = tr;
	FilePos = -1;
	ppp = 1;
}

// The Destructor
CDesignFileSymbol::~CDesignFileSymbol()
{
	drawingIterator it = m_method.begin();
	while (it != m_method.end()) 
	{
		delete *it;
		++ it;
	}
}
	

// Save this symbol into an archive
void CDesignFileSymbol::SaveXML(CXMLWriter &xml)
{
	xml.addTag( "NAME" ); 
	xml.addAttribute( "type", name_type );
	xml.addChildData( name );
	xml.closeTag();

	xml.addTag( "REF" );
	xml.addAttribute( "type", ref_type );
	xml.addChildData( reference );
	xml.closeTag();

	xml.addTag( "DESCRIPTION", description );
	xml.addTag( "PPP", ppp );
	
	for (int i = 0; i < fields.size(); i++)
	{
		fields[i].SaveXML( xml, false );
	}

	m_pDesign->SaveXML(xml, m_method, FALSE,FALSE,FALSE);
}


// Load this symbol from an archive
void CDesignFileSymbol::LoadXML(CTinyCadDoc *pDesign, CXMLReader &xml)
{
	int nt = 0;
	int rt = 0;
	CString name;
	m_pDesign = pDesign;

	while (xml.nextTag( name ))
	{
		if (name == "NAME")
		{
			xml.getChildData( name );
			xml.getAttribute( "type", nt );
			name_type = static_cast<SymbolFieldType>(nt);
		}
		else if (name == "REF")
		{
			xml.getChildData( reference );
			xml.getAttribute( "type", rt );
			ref_type = static_cast<SymbolFieldType>(rt);
		}
		else if (name == "DESCRIPTION")
		{ 
			xml.getChildData( description );
		}
		else if (name == "PPP")
		{
			xml.getChildData( ppp );
		}
		else if (name == "FIELD")
		{
			CSymbolField f;
			xml.intoTag();			
			f.LoadXML( xml );
			xml.outofTag();
			fields.push_back( f );
		}
		else if (name == "TinyCAD")
		{
			m_pDesign->ReadFileXML(xml, FALSE, m_method, TRUE );
			PostLoad(m_pDesign, m_Tr, m_method);
		}
	}
}



// Load this symbol from an archive
void CDesignFileSymbol::Load(CTinyCadDoc *pDesign, CStream &theArchive)
{
	CString dummy;
	theArchive >> name >> description >> reference >> ppp >> dummy; 

	DWORD number_of_fields = 0;
	if (!dummy.IsEmpty())
	{
		number_of_fields = atoi( dummy );
	}

    DWORD flags = number_of_fields >> 16;
    number_of_fields = number_of_fields & 0xffff;

	for (int i = 0; i < number_of_fields; i++)
	{
	  CSymbolField sf;
	  sf.Load( theArchive );
	  fields.push_back( sf );
	}

	name_type = static_cast<SymbolFieldType>(flags >> 3);
	ref_type = static_cast<SymbolFieldType>(flags & 0x7);


	m_pDesign = pDesign;
	LoadSymbol( pDesign, theArchive, m_Tr, m_method );
}



// Return the methods object
BOOL CDesignFileSymbol::GetMethod( CPoint &tr, drawingCollection &drawing )
{
	tr = m_Tr;
	drawing = m_method;

	return TRUE;
}



