/*
	TinyCAD program for schematic capture
	Copyright 1994-2004 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// XMLWriter.cpp: implementation of the CXMLWriter class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "tinycad.h"
#include "XMLWriter.h"
#include "XMLException.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CXMLWriter::CXMLWriter(CStream*	pOutput)
{
	m_child_data = false;
	m_pOutput = pOutput;
}

CXMLWriter::~CXMLWriter()
{

}


//////////////////////////////////////////////////////////////////////
// Add comments to the file
//////////////////////////////////////////////////////////////////////

void CXMLWriter::addComment( const char *comment )
{
	SendString("\n<!--");
	SendString(comment);
	SendString("-->\n");

}

//////////////////////////////////////////////////////////////////////
// Create a new tag 
//////////////////////////////////////////////////////////////////////


void CXMLWriter::addTag(const char *tag_name)
{
	// Do we need to close the previous tag?
	if (!m_child_data && m_tags.size() != 0)
	{
		closeOpenTag();
	}

	// Start a new tag, and push it's name onto the tag stack...
	m_tags.push_front( tag_name );

	SendString("<");
	SendString(tag_name);
	m_child_data = false;
}

void CXMLWriter::internal_addTag(const char *name, const char *data)
{
	addTag( name );
	internal_addChildData( data );
	closeTag();
}


//////////////////////////////////////////////////////////////////////
// Add attributes to the tag
//////////////////////////////////////////////////////////////////////


void CXMLWriter::internal_addAttribute(const char *name, const char *value)
{
	// Do we have child data?
	if (m_child_data)
	{
		throw new CXMLException(ERR_XML_HAVE_CHILD_DATA, TRUE);
	}

	// Is there an open tag?
	if (m_tags.size() == 0)
	{
		throw new CXMLException(ERR_XML_NO_TAG, TRUE);
	}

	SendString(" ");
	SendString(name);
	SendString("='");
	SendString(value);
	SendString("'");
}



//////////////////////////////////////////////////////////////////////
// Add child data to the current tags
//////////////////////////////////////////////////////////////////////


void CXMLWriter::internal_addChildData(const char *data)
{
	// Is there an open tag?
	if (m_tags.size() == 0)
	{
		throw new CXMLException(ERR_XML_NO_TAG, TRUE);		
	}

	if (!m_child_data)
	{
		closeOpenTag();
	}
	m_child_data = true;
	SendString( data );
}



//////////////////////////////////////////////////////////////////////
// Close the current tag
//////////////////////////////////////////////////////////////////////


void CXMLWriter::closeTag()
{
	// Did we have any child data?
	if (m_child_data)
	{
		// Yes, so need to close the tag with another tag
		SendString( "</" );
		SendString( m_tags.front() );
		SendString( ">\n" );
	}
	else
	{
		// No, so just close in a single tag
		SendString( "/>\n");
	}

	m_tags.pop_front();
	m_child_data = m_tags.size() != 0;
}


void CXMLWriter::closeOpenTag()
{
	SendString( ">" );
}



//
// Here are our conversion utilities to convert the data
// into XML strings
//
CString	CXMLWriter::makeString( int data )
{
	CString r;
	r.Format( "%d", data );
	return r;
}

CString	CXMLWriter::makeString( COLORREF data )
{
	CString r;
	r.Format( "%06X", data );
	return r;
}


CString	CXMLWriter::makeString( double data )
{
	CString r;
	r.Format("%lg", data );
	return r;
}

CString	CXMLWriter::makeString( CDPoint data )
{
	CString r;
	r.Format("%lg,%lg",data.x, data.y);
	return r;
}

CString	CXMLWriter::makeString( UINT data )
{
	CString r;
	r.Format("%u", data );
	return r;
}

CString	CXMLWriter::makeString( const char *data )
{
	CString r;

	r = data;
	int i = 0;
	int l = strlen( data );
	for (int p = 0; p < l; p++)
	{
		CString ins;

		switch (data[p])
		{
		case'<': 
			ins = "&lt;";
			break;
		case'>': 
			ins = "&gt;";
			break;
		case'&': 
			ins = "&amp;";
			break;
		case '\'': 
			ins = "&#039;";
			break;
		case '"': 
			ins = "&quot;";
			break;
		}

		if (!ins.IsEmpty())
		{
			r = r.Left(i) + ins + r.Mid(i+1);
			i += ins.GetLength();
		}
		else
		{
			i ++;
		}
	}

	return r;
}




// Send a completed string to the output
//
void CXMLWriter::SendString( const char *str)
{
	m_pOutput->Write( str, strlen( str ) );
}



// ENC is the basic 1 character encoding function to make 
// a char printable
#define ENC(c) (((c) & 077) + ' ')

//
// Add child data but uuencode it first...
//
void CXMLWriter::addChildDataUUencode( BYTE *data, UINT size )
{
	// Is there an open tag?
	if (m_tags.size() == 0)
	{
		throw new CXMLException(ERR_XML_NO_TAG, TRUE);
	}

	if (!m_child_data)
	{
		closeOpenTag();
	}

	m_child_data = true;

	addTag( "UUENCODE" );
	addAttribute( "size", size );
	closeOpenTag();
	m_child_data = true;
	SendString("\n");
	
	// Now send the data...
	UINT i = 0;
	while ( i < size )
	{
		char c[256];
		
		// 1 (up to) 45 character line
		int line_size = min(size - i, 45);
		int p = 0;
		c[p ++] = ENC(line_size);

		while (line_size > 0)
		{
			int i0 = data[i];
			i ++;

			int i1 = 0;
			if (i < size)
			{
				i1 = data[i];
				i ++;
			}

			int i2 = 0;
			if (i < size)
			{
				i2 = data[i];
				i ++;
			}

			// output one group of 3 bytes as 4 chars
			c[p ++] = ENC( i0 >> 2 );
			c[p ++] = ENC( (i0 << 4) & 060 | (i1 >> 4) & 017 );
			c[p ++] = ENC( (i1 << 2) & 074 | (i2 >> 6) & 03 );
			c[p ++] = ENC( i2 & 077 );
			
			line_size -= 3;
		}

		c[p ++] = '\n';
		c[p ++] = 0;

		SendString( makeString( c ) );
	}

	closeTag();
}

